__author__ = "sibirrer"

import numpy.testing as npt
import numpy as np
import pytest
from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Profiles.curved_arc_spp import CurvedArcSPP
from lenstronomy.LensModel.Profiles.curved_arc_tan_diff import CurvedArcTanDiff
from lenstronomy.Util import util
from lenstronomy.Util import param_util


class TestArcDistortions(object):
    """Tests the source model routines."""

    def setup_method(self):
        pass

    def test_radial_tangential_distortions(self):
        lens_model_list = ["CURVED_ARC_SPP", "SHEAR", "FLEXION"]
        center_x, center_y = 0.01, 0
        curvature = 1.0 / 2
        lens = LensModel(lens_model_list=lens_model_list)
        kwargs_lens = [
            {
                "tangential_stretch": 10,
                "radial_stretch": 1.0,
                "curvature": curvature,
                "direction": -10,
                "center_x": center_x,
                "center_y": center_y,
            },
            {"gamma1": -0.0, "gamma2": -0.0},
            {"g1": 0.0, "g2": 0.0, "g3": -0.0, "g4": 0},
        ]

        extensions = LensModelExtensions(lensModel=lens)

        (
            lambda_rad,
            lambda_tan,
            orientation_angle,
            dlambda_tan_dtan,
            dlambda_tan_drad,
            dlambda_rad_drad,
            dlambda_rad_dtan,
            dphi_tan_dtan,
            dphi_tan_drad,
            dphi_rad_drad,
            dphi_rad_dtan,
        ) = extensions.radial_tangential_differentials(
            x=center_x, y=center_y, kwargs_lens=kwargs_lens, smoothing_3rd=0.0001
        )
        print(orientation_angle, "orientation angle")
        l = 1.0 / dphi_tan_dtan
        npt.assert_almost_equal(l, 1.0 / curvature, decimal=6)

    def test_hessian_eigenvector_mp(self):
        lens_model_list = ["SIS", "SHEAR"]
        lens_mp = LensModel(
            lens_model_list=lens_model_list,
            lens_redshift_list=[0.5, 0.4],
            multi_plane=True,
            z_source=2,
        )
        lens = LensModel(lens_model_list=lens_model_list)
        x0, y0 = 1.0, 1.0
        kwargs_lens = [
            {"theta_E": 1, "center_x": 0, "center_y": 0},
            {"gamma1": 0.0, "gamma2": 0.00001},
        ]

        extensions = LensModelExtensions(lensModel=lens)
        extensions_mp = LensModelExtensions(lensModel=lens_mp)
        w0, w1, v11, v12, v21, v22 = extensions.hessian_eigenvectors(
            x0, y0, kwargs_lens, diff=None
        )
        (
            w0_mp,
            w1_mp,
            v11_mp,
            v12_mp,
            v21_mp,
            v22_mp,
        ) = extensions_mp.hessian_eigenvectors(x0, y0, kwargs_lens, diff=None)
        npt.assert_almost_equal(w0, w0_mp, decimal=3)
        npt.assert_almost_equal(w1, w1_mp, decimal=3)
        npt.assert_almost_equal(v11, v11_mp, decimal=3)
        npt.assert_almost_equal(v12, v12_mp, decimal=3)
        npt.assert_almost_equal(v21, v21_mp, decimal=3)
        npt.assert_almost_equal(v22, v22_mp, decimal=3)

    def test_radial_tangential_stretch(self):
        lens_model_list = ["SIS", "SHEAR"]
        lens_mp = LensModel(
            lens_model_list=lens_model_list,
            lens_redshift_list=[0.5, 0.4],
            multi_plane=True,
            z_source=2,
        )
        lens = LensModel(lens_model_list=lens_model_list)
        x0, y0 = 1.0, 1.0

        kwargs_lens = [
            {"theta_E": 1, "center_x": 0, "center_y": 0},
            {"gamma1": 0.0, "gamma2": 0.00001},
        ]

        extensions = LensModelExtensions(lensModel=lens)
        extensions_mp = LensModelExtensions(lensModel=lens_mp)

        (
            radial_stretch,
            tangential_stretch,
            v_rad1,
            v_rad2,
            v_tang1,
            v_tang2,
        ) = extensions.radial_tangential_stretch(x0, y0, kwargs_lens, diff=None)
        (
            radial_stretch_mp,
            tangential_stretch_mp,
            v_rad1_mp,
            v_rad2_mp,
            v_tang1_mp,
            v_tang2_mp,
        ) = extensions_mp.radial_tangential_stretch(x0, y0, kwargs_lens, diff=None)
        npt.assert_almost_equal(radial_stretch, radial_stretch_mp, decimal=4)
        npt.assert_almost_equal(tangential_stretch, tangential_stretch_mp, decimal=4)
        npt.assert_almost_equal(v_rad1, v_rad1_mp, decimal=4)
        npt.assert_almost_equal(v_rad2, v_rad2_mp, decimal=4)
        npt.assert_almost_equal(v_tang1, v_tang1_mp, decimal=4)
        npt.assert_almost_equal(v_tang2, v_tang2_mp, decimal=4)

    def test_radial_tangential_distortions_multi_plane(self):
        lens_model_list = ["SIS", "SHEAR"]
        lens_mp = LensModel(
            lens_model_list=lens_model_list,
            lens_redshift_list=[0.5, 0.4],
            multi_plane=True,
            z_source=2,
        )
        lens = LensModel(lens_model_list=lens_model_list)
        x0, y0 = 2.0, 1.0

        kwargs_lens = [
            {"theta_E": 1, "center_x": 0, "center_y": 0},
            {"gamma1": 0.0, "gamma2": 0.00001},
        ]

        extensions = LensModelExtensions(lensModel=lens)
        extensions_mp = LensModelExtensions(lensModel=lens_mp)

        (
            lambda_rad,
            lambda_tan,
            orientation_angle,
            dlambda_tan_dtan,
            dlambda_tan_drad,
            dlambda_rad_drad,
            dlambda_rad_dtan,
            dphi_tan_dtan,
            dphi_tan_drad,
            dphi_rad_drad,
            dphi_rad_dtan,
        ) = extensions.radial_tangential_differentials(
            x=x0, y=y0, kwargs_lens=kwargs_lens, smoothing_3rd=0.0001
        )

        (
            lambda_rad_mp,
            lambda_tan_mp,
            orientation_angle_mp,
            dlambda_tan_dtan_mp,
            dlambda_tan_drad_mp,
            dlambda_rad_drad_mp,
            dlambda_rad_dtan_mp,
            dphi_tan_dtan_mp,
            dphi_tan_drad_mp,
            dphi_rad_drad_mp,
            dphi_rad_dtan_mp,
        ) = extensions_mp.radial_tangential_differentials(
            x=x0, y=y0, kwargs_lens=kwargs_lens, smoothing_3rd=0.0001
        )

        npt.assert_almost_equal(lambda_rad, lambda_rad_mp, decimal=3)
        npt.assert_almost_equal(lambda_tan, lambda_tan_mp, decimal=3)
        npt.assert_almost_equal(dphi_tan_dtan, dphi_rad_dtan_mp, decimal=3)

    def test_radial_tangential_differentials(self):
        from lenstronomy.Util import util

        x, y = util.make_grid(numPix=10, deltapix=1)

        lens_model_list = ["SIS"]
        center_x, center_y = 0, 0
        lens = LensModel(lens_model_list=lens_model_list)
        kwargs_lens = [{"theta_E": 1, "center_x": center_x, "center_y": center_y}]

        extensions = LensModelExtensions(lensModel=lens)
        (
            lambda_rad,
            lambda_tan,
            orientation_angle,
            dlambda_tan_dtan,
            dlambda_tan_drad,
            dlambda_rad_drad,
            dlambda_rad_dtan,
            dphi_tan_dtan,
            dphi_tan_drad,
            dphi_rad_drad,
            dphi_rad_dtan,
        ) = extensions.radial_tangential_differentials(
            2, 2, kwargs_lens, smoothing_3rd=0.001
        )

        npt.assert_almost_equal(lambda_rad, 1, decimal=5)
        npt.assert_almost_equal(lambda_tan, 1.5469181606780271, decimal=5)

        (
            lambda_rad,
            lambda_tan,
            orientation_angle,
            dlambda_tan_dtan,
            dlambda_tan_drad,
            dlambda_rad_drad,
            dlambda_rad_dtan,
            dphi_tan_dtan,
            dphi_tan_drad,
            dphi_rad_drad,
            dphi_rad_dtan,
        ) = extensions.radial_tangential_differentials(
            np.array([2]), np.array([2]), kwargs_lens, smoothing_3rd=0.001
        )

        npt.assert_almost_equal(lambda_rad, 1, decimal=5)
        npt.assert_almost_equal(lambda_tan, 1.5469181606780271, decimal=5)

        mag = lens.magnification(x, y, kwargs_lens)
        (
            lambda_rad,
            lambda_tan,
            orientation_angle,
            dlambda_tan_dtan,
            dlambda_tan_drad,
            dlambda_rad_drad,
            dlambda_rad_dtan,
            dphi_tan_dtan,
            dphi_tan_drad,
            dphi_rad_drad,
            dphi_rad_dtan,
        ) = extensions.radial_tangential_differentials(
            x, y, kwargs_lens, smoothing_3rd=0.001
        )
        mag_tang_rad = lambda_tan * lambda_rad
        npt.assert_almost_equal(mag_tang_rad, mag, decimal=5)

    def test_curved_arc_estimate(self):
        lens_model_list = ["SPP"]
        lens = LensModel(lens_model_list=lens_model_list)
        arc = LensModel(lens_model_list=["CURVED_ARC_SPP"])
        theta_E = 4
        gamma = 2.0
        kwargs_lens = [
            {"theta_E": theta_E, "gamma": gamma, "center_x": 0, "center_y": 0}
        ]
        ext = LensModelExtensions(lensModel=lens)
        x_0, y_0 = 5, 0
        kwargs_arc = ext.curved_arc_estimate(x_0, y_0, kwargs_lens)
        (
            theta_E_arc,
            gamma_arc,
            center_x_spp_arc,
            center_y_spp_arc,
        ) = CurvedArcSPP.stretch2spp(**kwargs_arc)
        npt.assert_almost_equal(theta_E_arc, theta_E, decimal=4)
        npt.assert_almost_equal(gamma_arc, gamma, decimal=3)
        npt.assert_almost_equal(center_x_spp_arc, 0, decimal=3)
        npt.assert_almost_equal(center_y_spp_arc, 0, decimal=3)
        x, y = util.make_grid(numPix=10, deltapix=1)
        alpha_x, alpha_y = lens.alpha(x, y, kwargs_lens)
        alpha0_x, alpha0_y = lens.alpha(x_0, y_0, kwargs_lens)
        alpha_x_arc, alpha_y_arc = arc.alpha(x, y, [kwargs_arc])
        npt.assert_almost_equal(alpha_x_arc, alpha_x - alpha0_x, decimal=3)
        npt.assert_almost_equal(alpha_y_arc, alpha_y - alpha0_y, decimal=3)

        x_0, y_0 = 0.0, 3
        kwargs_arc = ext.curved_arc_estimate(x_0, y_0, kwargs_lens)
        (
            theta_E_arc,
            gamma_arc,
            center_x_spp_arc,
            center_y_spp_arc,
        ) = CurvedArcSPP.stretch2spp(**kwargs_arc)
        print(kwargs_arc)
        print(theta_E_arc, gamma_arc, center_x_spp_arc, center_y_spp_arc)
        npt.assert_almost_equal(theta_E_arc, theta_E, decimal=4)
        npt.assert_almost_equal(gamma_arc, gamma, decimal=3)
        npt.assert_almost_equal(center_x_spp_arc, 0, decimal=3)
        npt.assert_almost_equal(center_y_spp_arc, 0, decimal=3)

        x_0, y_0 = -2, -3
        kwargs_arc = ext.curved_arc_estimate(x_0, y_0, kwargs_lens)
        (
            theta_E_arc,
            gamma_arc,
            center_x_spp_arc,
            center_y_spp_arc,
        ) = CurvedArcSPP.stretch2spp(**kwargs_arc)
        npt.assert_almost_equal(theta_E_arc, theta_E, decimal=4)
        npt.assert_almost_equal(gamma_arc, gamma, decimal=3)
        npt.assert_almost_equal(center_x_spp_arc, 0, decimal=3)
        npt.assert_almost_equal(center_y_spp_arc, 0, decimal=3)

    def test_curved_arc_finite_area(self):
        lens_model_list = ["SPP"]
        lens = LensModel(lens_model_list=lens_model_list)
        arc = LensModel(lens_model_list=["CURVED_ARC_SPP"])
        theta_E = 4
        gamma = 2.0
        kwargs_lens = [
            {"theta_E": theta_E, "gamma": gamma, "center_x": 0, "center_y": 0}
        ]
        ext = LensModelExtensions(lensModel=lens)
        x_0, y_0 = 5, 0
        kwargs_arc = ext.curved_arc_estimate(x_0, y_0, kwargs_lens)

        dr = 0.001
        kwargs_arc_finite = ext.curved_arc_finite_area(x_0, y_0, kwargs_lens, dr)
        npt.assert_almost_equal(
            kwargs_arc["direction"], kwargs_arc_finite["direction"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_arc["radial_stretch"], kwargs_arc_finite["radial_stretch"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_arc["tangential_stretch"],
            kwargs_arc_finite["tangential_stretch"],
            decimal=3,
        )
        npt.assert_almost_equal(
            kwargs_arc["curvature"], kwargs_arc_finite["curvature"], decimal=3
        )

    def test_curved_arc_estimate_tan_diff(self):
        arc_tan_diff = CurvedArcTanDiff()
        lens_model_list = ["SIE"]
        lens = LensModel(lens_model_list=lens_model_list)
        arc = LensModel(lens_model_list=["CURVED_ARC_TAN_DIFF"])
        theta_E = 4

        # here we model an off-axis ellisoid relative to the x-axis
        e1, e2 = 0.0, -0.1
        x_0, y_0 = 5, 0
        kwargs_lens = [
            {"theta_E": theta_E, "e1": e1, "e2": e2, "center_x": 0, "center_y": 0}
        ]
        ext = LensModelExtensions(lensModel=lens)

        kwargs_arc = ext.curved_arc_estimate(
            x_0, y_0, kwargs_lens, tan_diff=True, smoothing_3rd=0.01
        )
        (
            theta_E_sie,
            e1_sie,
            e2_sie,
            kappa_ext,
            center_x_sis,
            center_y_sis,
        ) = arc_tan_diff.stretch2sie_mst(**kwargs_arc)
        print(theta_E_sie, e1_sie, e2_sie, center_x_sis, center_y_sis)
        npt.assert_almost_equal(e2_sie - e2, 0, decimal=1)
        npt.assert_almost_equal(e1_sie, e1, decimal=3)

        # here we model an off-axis ellisoid relative to the y-axis
        e1, e2 = 0.1, 0.0
        x_0, y_0 = 0, 5
        kwargs_lens = [
            {"theta_E": theta_E, "e1": e1, "e2": e2, "center_x": 0, "center_y": 0}
        ]
        ext = LensModelExtensions(lensModel=lens)

        kwargs_arc = ext.curved_arc_estimate(
            x_0, y_0, kwargs_lens, tan_diff=True, smoothing_3rd=0.01
        )
        (
            theta_E_sie,
            e1_sie,
            e2_sie,
            kappa_ext,
            center_x_sis,
            center_y_sis,
        ) = arc_tan_diff.stretch2sie_mst(**kwargs_arc)
        print(theta_E_sie, e1_sie, e2_sie, center_x_sis, center_y_sis)
        npt.assert_almost_equal(e1_sie - e1, 0, decimal=1)
        npt.assert_almost_equal(e2_sie, e2, decimal=3)

        x, y = util.make_grid(numPix=100, deltapix=0.1)
        kappa = lens.kappa(x, y, kwargs_lens)
        kappa_arc = arc.kappa(x, y, [kwargs_arc])

    def test_arcs_at_image_position(self):
        # lensing quantities
        kwargs_spp = {
            "theta_E": 1.26,
            "gamma": 2.0,
            "e1": 0.1,
            "e2": -0.1,
            "center_x": 0.0,
            "center_y": 0.0,
        }  # parameters of the deflector lens model

        # the lens model is a supperposition of an elliptical lens model with external shear
        lens_model_list = ["SPEP"]  # , 'SHEAR']
        kwargs_lens = [kwargs_spp]  # , kwargs_shear]
        lens_model_class = LensModel(lens_model_list=lens_model_list)
        lensEquationSolver = LensEquationSolver(lens_model_class)
        source_x = 0.0
        source_y = 0.05
        x_image, y_image = lensEquationSolver.findBrightImage(
            source_x,
            source_y,
            kwargs_lens,
            numImages=4,
            min_distance=0.05,
            search_window=5,
        )
        arc_model = LensModel(lens_model_list=["CURVED_ARC_SPP", "SHIFT"])
        for i in range(len(x_image)):
            x0, y0 = x_image[i], y_image[i]
            print(x0, y0, i)
            ext = LensModelExtensions(lensModel=lens_model_class)
            kwargs_arc_i = ext.curved_arc_estimate(x0, y0, kwargs_lens)
            alpha_x, alpha_y = lens_model_class.alpha(x0, y0, kwargs_lens)
            kwargs_arc = [kwargs_arc_i, {"alpha_x": alpha_x, "alpha_y": alpha_y}]
            print(kwargs_arc_i)
            direction = kwargs_arc_i["direction"]
            print(np.cos(direction), np.sin(direction))
            x, y = util.make_grid(numPix=5, deltapix=0.01)
            x = x0
            y = y0
            gamma1_arc, gamma2_arc = arc_model.gamma(x, y, kwargs_arc)
            gamma1, gamma2 = lens_model_class.gamma(x, y, kwargs_lens)
            print(gamma1, gamma2)
            npt.assert_almost_equal(gamma1_arc, gamma1, decimal=3)
            npt.assert_almost_equal(gamma2_arc, gamma2, decimal=3)
            (
                theta_E_arc,
                gamma_arc,
                center_x_spp_arc,
                center_y_spp_arc,
            ) = CurvedArcSPP.stretch2spp(**kwargs_arc_i)
            print(theta_E_arc, gamma_arc, center_x_spp_arc, center_y_spp_arc)
            npt.assert_almost_equal(center_x_spp_arc, 0, decimal=3)
            npt.assert_almost_equal(center_y_spp_arc, 0, decimal=3)

    def test_analytic_differentials_spp(self):
        """Test the analytical differentials of the spherical power-law mass profile.

        :return:
        """
        lens_model_class = LensModel(lens_model_list=["SPP"])
        theta_E = 2
        gamma = 2.2
        kwargs_lens = [
            {"theta_E": theta_E, "gamma": gamma, "center_x": 0, "center_y": 0}
        ]
        ext = LensModelExtensions(lensModel=lens_model_class)
        # we leave out the very high magnification region where the numerical differentials might not be as accurate as required in the tests
        x = np.append(
            np.linspace(start=0.1, stop=theta_E - 0.1, num=9),
            np.linspace(start=theta_E + 0.1, stop=3 * theta_E, num=9),
        )
        y = np.zeros_like(x)
        (
            lambda_rad,
            lambda_tan,
            orientation_angle,
            dlambda_tan_dtan,
            dlambda_tan_drad,
            dlambda_rad_drad,
            dlambda_rad_dtan,
            dphi_tan_dtan,
            dphi_tan_drad,
            dphi_rad_drad,
            dphi_rad_dtan,
        ) = ext.radial_tangential_differentials(x, y, kwargs_lens)

        def _lambda_t_analytic(r, theta_E, gamma):
            """Analytic expression for lambda_tan.

            :param r: radius
            :return:
            """
            return (1 - (theta_E / r) ** (gamma - 1)) ** -1

        lambda_tan_analytic = _lambda_t_analytic(x, theta_E, gamma)
        npt.assert_almost_equal(lambda_tan_analytic, lambda_tan, decimal=5)

        def _lambda_r_analytic(r, theta_E, gamma):
            """Analytic expression for lambda_rad.

            :param r: radius
            :return:
            """
            return (1 + (gamma - 2) * (theta_E / r) ** (gamma - 1)) ** -1

        lambda_rad_analytic = _lambda_r_analytic(x, theta_E, gamma)
        npt.assert_almost_equal(lambda_rad_analytic, lambda_rad, decimal=5)

        def _lambda_t_dr_analytic(r, theta_E, gamma):
            """Analytic expression for  d(lambda_tan) / dr.

            :param r: radius
            :return:
            """
            return (
                (1 - gamma)
                * (theta_E / r) ** gamma
                / (theta_E * (1 - (theta_E / r) ** (gamma - 1)) ** 2)
            )

        dlambda_tan_drad_analytic = _lambda_t_dr_analytic(x, theta_E, gamma)
        npt.assert_almost_equal(
            dlambda_tan_drad_analytic / dlambda_tan_drad, 1, decimal=2
        )

        def _lambda_r_dr_analytic(r, theta_E, gamma):
            """Analytic expression for  d(lambda_tan) / dr.

            :param r: radius
            :return:
            """
            return (
                (1 - gamma)
                * (2 - gamma)
                * (theta_E / r) ** gamma
                / (theta_E * (1 + (theta_E / r) ** (gamma - 1) * (gamma - 2)) ** 2)
            )

        dlambda_r_drad_analytic = _lambda_r_dr_analytic(x, theta_E, gamma)
        npt.assert_almost_equal(
            dlambda_r_drad_analytic / dlambda_rad_drad, 1, decimal=2
        )

    def test_analytic_differential_pemd(self):
        lens_model_class = LensModel(lens_model_list=["EPL"])
        theta_E = 4
        gamma = 2.0

        ext = LensModelExtensions(lensModel=lens_model_class)
        # we leave out the very high magnification region where the numerical differentials might not be as accurate as required in the tests
        x = np.append(
            np.linspace(start=0.2 * theta_E, stop=theta_E * (1 - 0.2), num=9),
            np.linspace(start=theta_E * (1 + 0.5), stop=3 * theta_E, num=9),
        )
        y = np.zeros_like(x)

        def _dlambda_t_dr_analytic(r, theta_E, gamma):
            """Analytic expression for  d(lambda_tan) / dr.

            :param r: circularized radius
            :return:
            """
            return (
                (1 - gamma)
                * (theta_E / r) ** gamma
                / (theta_E * (1 - (theta_E / r) ** (gamma - 1)) ** 2)
            )

        def _dlambda_t_dt_analytic(x, y, theta_E, gamma, q, phi_G):
            # polar coordinates with respect to x-axis
            r, phi = param_util.cart2polar(x, y, center_x=0, center_y=0)
            epsilon = (1 - q**2) / (1 + q**2)

            # radial component in respect to rotation of deflector
            r_ = r * np.sqrt(1 - epsilon * np.cos(2 * (phi - phi_G)))
            # equivalent  Einstein radius for elliptical mass definition
            theta_E_prim = np.sqrt(2 * q / (1 + q**2)) * theta_E
            dlambda_t_dr = _dlambda_t_dr_analytic(r_, theta_E_prim, gamma)
            dr_de_t = (
                epsilon
                * np.sin(2 * (phi - phi_G))
                / np.sqrt(1 - epsilon * np.cos(2 * (phi - phi_G)))
            )
            return dlambda_t_dr * dr_de_t, dlambda_t_dr

        # define ellipticity with axis ratio and orientation

        # ellipticity aligned -> leading to zero differential
        phi_G, q = 0.0, 0.7
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = [
            {
                "theta_E": theta_E,
                "gamma": gamma,
                "e1": e1,
                "e2": e2,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        (
            lambda_rad,
            lambda_tan,
            orientation_angle,
            dlambda_tan_dtan,
            dlambda_tan_drad,
            dlambda_rad_drad,
            dlambda_rad_dtan,
            dphi_tan_dtan,
            dphi_tan_drad,
            dphi_rad_drad,
            dphi_rad_dtan,
        ) = ext.radial_tangential_differentials(x, y, kwargs_lens)
        dlambda_tan_dtan_analytic, dlambda_t_dr_analtic = _dlambda_t_dt_analytic(
            x, y, theta_E, gamma, q, phi_G
        )
        npt.assert_almost_equal(dlambda_tan_dtan_analytic, 0, decimal=5)
        npt.assert_almost_equal(dlambda_tan_dtan, 0, decimal=1)

        # ellipticity aligned -> leading to zero differential
        phi_G, q = np.pi / 2, 0.7
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = [
            {
                "theta_E": theta_E,
                "gamma": gamma,
                "e1": e1,
                "e2": e2,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        (
            lambda_rad,
            lambda_tan,
            orientation_angle,
            dlambda_tan_dtan,
            dlambda_tan_drad,
            dlambda_rad_drad,
            dlambda_rad_dtan,
            dphi_tan_dtan,
            dphi_tan_drad,
            dphi_rad_drad,
            dphi_rad_dtan,
        ) = ext.radial_tangential_differentials(x, y, kwargs_lens)
        dlambda_tan_dtan_analytic, dlambda_t_dr_analtic = _dlambda_t_dt_analytic(
            x, y, theta_E, gamma, q, phi_G
        )
        npt.assert_almost_equal(dlambda_tan_dtan_analytic, 0, decimal=5)
        npt.assert_almost_equal(dlambda_tan_dtan, 0, decimal=1)

        phi_G, q = np.pi / 4, 0.7
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = [
            {
                "theta_E": theta_E,
                "gamma": gamma,
                "e1": e1,
                "e2": e2,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        (
            lambda_rad,
            lambda_tan,
            orientation_angle,
            dlambda_tan_dtan,
            dlambda_tan_drad,
            dlambda_rad_drad,
            dlambda_rad_dtan,
            dphi_tan_dtan,
            dphi_tan_drad,
            dphi_rad_drad,
            dphi_rad_dtan,
        ) = ext.radial_tangential_differentials(x, y, kwargs_lens)
        dlambda_tan_dtan_analytic, dlambda_t_dr_analtic = _dlambda_t_dt_analytic(
            x, y, theta_E, gamma, q, phi_G
        )
        print(dlambda_rad_drad, "dlambda_rad")
        npt.assert_almost_equal(dlambda_tan_drad / dlambda_t_dr_analtic, 1, decimal=2)
        print(dlambda_tan_dtan_analytic / dlambda_tan_dtan, "test")
        npt.assert_almost_equal(
            dlambda_tan_dtan_analytic / dlambda_tan_dtan, 1, decimal=2
        )

        phi_G, q = np.pi / 8, 0.7
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = [
            {
                "theta_E": theta_E,
                "gamma": gamma,
                "e1": e1,
                "e2": e2,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        (
            lambda_rad,
            lambda_tan,
            orientation_angle,
            dlambda_tan_dtan,
            dlambda_tan_drad,
            dlambda_rad_drad,
            dlambda_rad_dtan,
            dphi_tan_dtan,
            dphi_tan_drad,
            dphi_rad_drad,
            dphi_rad_dtan,
        ) = ext.radial_tangential_differentials(x, y, kwargs_lens)
        dlambda_tan_dtan_analytic, dlambda_t_dr_analtic = _dlambda_t_dt_analytic(
            x, y, theta_E, gamma, q, phi_G
        )
        print(dlambda_tan_dtan_analytic / dlambda_tan_dtan, "test")
        npt.assert_almost_equal(
            dlambda_tan_dtan_analytic / dlambda_tan_dtan, 1, decimal=2
        )

        phi_G, q = 0.3, 0.7
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = [
            {
                "theta_E": theta_E,
                "gamma": gamma,
                "e1": e1,
                "e2": e2,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        (
            lambda_rad,
            lambda_tan,
            orientation_angle,
            dlambda_tan_dtan,
            dlambda_tan_drad,
            dlambda_rad_drad,
            dlambda_rad_dtan,
            dphi_tan_dtan,
            dphi_tan_drad,
            dphi_rad_drad,
            dphi_rad_dtan,
        ) = ext.radial_tangential_differentials(x, y, kwargs_lens)
        dlambda_tan_dtan_analytic, dlambda_t_dr_analtic = _dlambda_t_dt_analytic(
            x, y, theta_E, gamma, q, phi_G
        )
        print(dlambda_tan_dtan_analytic / dlambda_tan_dtan, "test")
        npt.assert_almost_equal(
            dlambda_tan_dtan_analytic / dlambda_tan_dtan, 1, decimal=2
        )


if __name__ == "__main__":
    pytest.main("-k TestLensModel")



import copy
import scipy.integrate as integrate
import numpy as np

__all__ = ["ProfileIntegrals"]


class ProfileIntegrals(object):
    """
    class to perform integrals of spherical profiles to compute:
    - projected densities
    - enclosed densities
    - projected enclosed densities
    """

    def __init__(self, profile_class):
        """

        :param profile_class: list of lens models
        """
        self._profile = profile_class

    def mass_enclosed_3d(self, r, kwargs_profile, lens_param=False):
        """Computes the mass enclosed within a sphere of radius r.

        :param r: radius (arcsec)
        :param kwargs_profile: keyword argument list with lens model parameters
        :param lens_param: boolean, if True uses the lens model parameterization in
            computing the 3d density convention and the return is the convergence
        :return: 3d mass enclosed of r
        """
        kwargs = copy.deepcopy(kwargs_profile)
        kwargs.pop("center_x", None)
        kwargs.pop("center_y", None)
        # integral of self._profile.density(x)* 4*np.pi * x^2 *dx, 0,r
        if lens_param is True:
            out = integrate.quad(
                lambda x: self._profile.density_lens(x, **kwargs) * 4 * np.pi * x**2,
                0,
                r,
            )
        else:
            out = integrate.quad(
                lambda x: self._profile.density(x, **kwargs) * 4 * np.pi * x**2, 0, r
            )
        return out[0]

    def density_2d(self, r, kwargs_profile, lens_param=False):
        """Computes the projected density along the line-of-sight.

        :param r: radius (arcsec)
        :param kwargs_profile: keyword argument list with lens model parameters
        :param lens_param: boolean, if True uses the lens model parameterization in
            computing the 3d density convention and the return is the convergence
        :return: 2d projected density at projected radius r
        """
        kwargs = copy.deepcopy(kwargs_profile)
        kwargs.pop("center_x", None)
        kwargs.pop("center_y", None)
        # integral of self._profile.density(np.sqrt(x^2+r^2))* dx, 0, infty
        if lens_param is True:
            out = integrate.quad(
                lambda x: 2
                * self._profile.density_lens(np.sqrt(x**2 + r**2), **kwargs),
                0,
                100,
            )
        else:
            out = integrate.quad(
                lambda x: 2 * self._profile.density(np.sqrt(x**2 + r**2), **kwargs),
                0,
                100,
            )
        return out[0]

    def mass_enclosed_2d(self, r, kwargs_profile):
        """
        computes the mass enclosed the projected line-of-sight
        :param r: radius (arcsec)
        :param kwargs_profile: keyword argument list with lens model parameters
        :return: projected mass enclosed radius r
        """
        kwargs = copy.deepcopy(kwargs_profile)
        kwargs.pop("center_x", None)
        kwargs.pop("center_y", None)
        # integral of self.density_2d(x)* 2*np.pi * x *dx, 0, r
        out = integrate.quad(lambda x: self.density_2d(x, kwargs) * 2 * np.pi * x, 0, r)
        return out[0]



import pytest
import numpy.testing as npt
import numpy as np
import unittest

import lenstronomy.Util.util as util
import lenstronomy.Util.kernel_util as kernel_util
from lenstronomy.ImSim.image_model import ImageModel


class TestNumerics(object):
    def setup_method(self):
        # we define a model consisting of a singe Sersric profile
        from lenstronomy.LightModel.light_model import LightModel

        light_model_list = ["SERSIC_ELLIPSE"]
        self.lightModel = LightModel(light_model_list=light_model_list)
        self.kwargs_light = [
            {
                "amp": 100,
                "R_sersic": 0.5,
                "n_sersic": 3,
                "e1": 0,
                "e2": 0,
                "center_x": 0.02,
                "center_y": 0,
            }
        ]

        # we define a pixel grid and a higher resolution super sampling factor
        self._supersampling_factor = 5
        numPix = 61  # cutout pixel size
        deltaPix = 0.05  # pixel size in arcsec (area per pixel = deltaPix**2)
        (
            x,
            y,
            ra_at_xy_0,
            dec_at_xy_0,
            x_at_radec_0,
            y_at_radec_0,
            Mpix2coord,
            Mcoord2pix,
        ) = util.make_grid_with_coordtransform(
            numPix=numPix,
            deltapix=deltaPix,
            subgrid_res=1,
            left_lower=False,
            inverse=False,
        )
        flux = self.lightModel.surface_brightness(x, y, kwargs_list=self.kwargs_light)
        flux = util.array2image(flux)
        flux_max = np.max(flux)
        conv_pixels_partial = np.zeros((numPix, numPix), dtype=bool)
        conv_pixels_partial[flux >= flux_max / 20] = True
        self._conv_pixels_partial = conv_pixels_partial

        # high resolution ray-tracing and high resolution convolution, the full calculation
        self.kwargs_numerics_true = {
            "supersampling_factor": self._supersampling_factor,
            # super sampling factor of (partial) high resolution ray-tracing
            "compute_mode": "regular",  # 'regular' or 'adaptive'
            "supersampling_convolution": True,
            # bool, if True, performs the supersampled convolution (either on regular or adaptive grid)
            "supersampling_kernel_size": None,
            # size of the higher resolution kernel region (can be smaller than the original kernel). None leads to use the full size
            "flux_evaluate_indexes": None,  # bool mask, if None, it will evaluate all (sub) pixels
            "supersampled_indexes": None,
            # bool mask of pixels to be computed in supersampled grid (only for adaptive mode)
            "compute_indexes": None,
            # bool mask of pixels to be computed the PSF response (flux being added to). Only used for adaptive mode and can be set =likelihood mask.
            "point_source_supersampling_factor": 1,
            # int, supersampling factor when rendering a point source (not used in this script)
        }

        # high resolution convolution on a smaller PSF with low resolution convolution on the edges of the PSF and high resolution ray tracing
        self.kwargs_numerics_high_res_narrow = {
            "supersampling_factor": self._supersampling_factor,
            "compute_mode": "regular",
            "supersampling_convolution": True,
            "supersampling_kernel_size": 5,
        }

        # low resolution convolution based on high resolution ray-tracing grid
        self.kwargs_numerics_low_conv_high_grid = {
            "supersampling_factor": self._supersampling_factor,
            "compute_mode": "regular",
            "supersampling_convolution": False,
            # does not matter for supersampling_factor=1
            "supersampling_kernel_size": None,
            # does not matter for supersampling_factor=1
        }

        # low resolution convolution with a subset of pixels with high resolution ray-tracing
        self.kwargs_numerics_low_conv_high_adaptive = {
            "supersampling_factor": self._supersampling_factor,
            "compute_mode": "adaptive",
            "supersampling_convolution": False,
            # does not matter for supersampling_factor=1
            "supersampling_kernel_size": None,
            # does not matter for supersampling_factor=1
            "supersampled_indexes": self._conv_pixels_partial,
            "convolution_kernel_size": 9,
        }

        # low resolution convolution with a subset of pixels with high resolution ray-tracing and high resoluton convolution on smaller kernel size
        self.kwargs_numerics_high_adaptive = {
            "supersampling_factor": self._supersampling_factor,
            "compute_mode": "adaptive",
            "supersampling_convolution": True,
            # does not matter for supersampling_factor=1
            "supersampling_kernel_size": 5,  # does not matter for supersampling_factor=1
            "supersampled_indexes": self._conv_pixels_partial,
            "convolution_kernel_size": 9,
        }

        # low resolution convolution and low resolution ray tracing, the simplest calculation
        self.kwargs_numerics_low_res = {
            "supersampling_factor": 1,
            "compute_mode": "regular",
            "supersampling_convolution": False,  # does not matter for supersampling_factor=1
            "supersampling_kernel_size": None,  # does not matter for supersampling_factor=1
            "convolution_kernel_size": 9,
        }

        flux_evaluate_indexes = np.zeros((numPix, numPix), dtype=bool)
        flux_evaluate_indexes[flux >= flux_max / 1000] = True
        # low resolution convolution on subframe
        self.kwargs_numerics_partial = {
            "supersampling_factor": 1,
            "compute_mode": "regular",
            "supersampling_convolution": False,
            # does not matter for supersampling_factor=1
            "supersampling_kernel_size": None,  # does not matter for supersampling_factor=1
            "flux_evaluate_indexes": flux_evaluate_indexes,
            "convolution_kernel_size": 9,
        }

        # import PSF file
        kernel_super = kernel_util.kernel_gaussian(
            num_pix=11 * self._supersampling_factor,
            delta_pix=deltaPix / self._supersampling_factor,
            fwhm=0.1,
        )

        kernel_size = 9
        kernel_super = kernel_util.cut_psf(
            psf_data=kernel_super, psf_size=kernel_size * self._supersampling_factor
        )

        # make instance of the PixelGrid class
        from lenstronomy.Data.pixel_grid import PixelGrid

        kwargs_grid = {
            "nx": numPix,
            "ny": numPix,
            "transform_pix2angle": Mpix2coord,
            "ra_at_xy_0": ra_at_xy_0,
            "dec_at_xy_0": dec_at_xy_0,
        }
        self.pixel_grid = PixelGrid(**kwargs_grid)

        # make instance of the PSF class
        from lenstronomy.Data.psf import PSF

        kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel_super,
            "point_source_supersampling_factor": self._supersampling_factor,
        }
        self.psf_class = PSF(**kwargs_psf)

        # without convolution
        image_model_true = ImageModel(
            self.pixel_grid,
            self.psf_class,
            lens_light_model_class=self.lightModel,
            kwargs_numerics=self.kwargs_numerics_true,
        )
        self.image_true = image_model_true.image(kwargs_lens_light=self.kwargs_light)

    def test_full(self):
        image_model_true = ImageModel(
            self.pixel_grid,
            self.psf_class,
            lens_light_model_class=self.lightModel,
            kwargs_numerics=self.kwargs_numerics_true,
        )
        image_unconvolved = image_model_true.image(
            kwargs_lens_light=self.kwargs_light, unconvolved=True
        )
        npt.assert_almost_equal(
            np.sum(self.image_true) / np.sum(image_unconvolved), 1, decimal=2
        )

    def test_high_res_narrow(self):
        image_model = ImageModel(
            self.pixel_grid,
            self.psf_class,
            lens_light_model_class=self.lightModel,
            kwargs_numerics=self.kwargs_numerics_high_res_narrow,
        )
        image_conv = image_model.image(
            kwargs_lens_light=self.kwargs_light, unconvolved=False
        )
        npt.assert_almost_equal(
            (self.image_true - image_conv) / self.image_true, 0, decimal=2
        )

    def test_low_conv_high_grid(self):
        image_model = ImageModel(
            self.pixel_grid,
            self.psf_class,
            lens_light_model_class=self.lightModel,
            kwargs_numerics=self.kwargs_numerics_low_conv_high_grid,
        )
        image_conv = image_model.image(
            kwargs_lens_light=self.kwargs_light, unconvolved=False
        )
        npt.assert_almost_equal(
            (self.image_true - image_conv) / self.image_true, 0, decimal=1
        )

    def test_low_conv_high_adaptive(self):
        image_model = ImageModel(
            self.pixel_grid,
            self.psf_class,
            lens_light_model_class=self.lightModel,
            kwargs_numerics=self.kwargs_numerics_low_conv_high_adaptive,
        )
        image_conv = image_model.image(
            kwargs_lens_light=self.kwargs_light, unconvolved=False
        )
        npt.assert_almost_equal(
            (self.image_true - image_conv) / self.image_true, 0, decimal=1
        )

    def test_high_adaptive(self):
        image_model = ImageModel(
            self.pixel_grid,
            self.psf_class,
            lens_light_model_class=self.lightModel,
            kwargs_numerics=self.kwargs_numerics_high_adaptive,
        )
        image_conv = image_model.image(
            kwargs_lens_light=self.kwargs_light, unconvolved=False
        )
        npt.assert_almost_equal(
            (self.image_true - image_conv) / self.image_true, 0, decimal=1
        )

    def test_low_res(self):
        image_model = ImageModel(
            self.pixel_grid,
            self.psf_class,
            lens_light_model_class=self.lightModel,
            kwargs_numerics=self.kwargs_numerics_low_res,
        )
        image_conv = image_model.image(
            kwargs_lens_light=self.kwargs_light, unconvolved=False
        )
        npt.assert_almost_equal(
            (self.image_true - image_conv) / self.image_true, 0, decimal=1
        )

    def test_sub_frame(self):
        image_model = ImageModel(
            self.pixel_grid,
            self.psf_class,
            lens_light_model_class=self.lightModel,
            kwargs_numerics=self.kwargs_numerics_partial,
        )
        image_conv = image_model.image(
            kwargs_lens_light=self.kwargs_light, unconvolved=False
        )
        delta = (self.image_true - image_conv) / self.image_true
        npt.assert_almost_equal(delta[self._conv_pixels_partial], 0, decimal=1)

    def test_property_access(self):
        image_model = ImageModel(
            self.pixel_grid,
            self.psf_class,
            lens_light_model_class=self.lightModel,
            kwargs_numerics=self.kwargs_numerics_true,
        )
        grid_supersampling_factor = image_model.ImageNumerics.grid_supersampling_factor
        assert grid_supersampling_factor == self._supersampling_factor

        kwargs_numerics = {
            "supersampling_factor": 1,
            "compute_mode": "regular",
            "supersampling_convolution": False,
        }
        image_model = ImageModel(
            self.pixel_grid,
            self.psf_class,
            lens_light_model_class=self.lightModel,
            kwargs_numerics=kwargs_numerics,
        )
        from lenstronomy.ImSim.Numerics.convolution import PixelKernelConvolution

        convolution_class = image_model.ImageNumerics.convolution_class
        assert isinstance(convolution_class, PixelKernelConvolution)

        kwargs_numerics = {
            "supersampling_factor": 2,
            "compute_mode": "regular",
            "supersampling_convolution": True,
        }
        image_model = ImageModel(
            self.pixel_grid,
            self.psf_class,
            lens_light_model_class=self.lightModel,
            kwargs_numerics=kwargs_numerics,
        )
        from lenstronomy.ImSim.Numerics.convolution import SubgridKernelConvolution

        convolution_class = image_model.ImageNumerics.convolution_class
        assert isinstance(convolution_class, SubgridKernelConvolution)

        kwargs_numerics = {
            "supersampling_factor": 2,
            "compute_mode": "adaptive",
            "supersampling_convolution": True,
        }
        image_model = ImageModel(
            self.pixel_grid,
            self.psf_class,
            lens_light_model_class=self.lightModel,
            kwargs_numerics=kwargs_numerics,
        )
        from lenstronomy.ImSim.Numerics.adaptive_numerics import AdaptiveConvolution

        convolution_class = image_model.ImageNumerics.convolution_class
        assert isinstance(convolution_class, AdaptiveConvolution)

        kwargs_numerics = {"compute_mode": "regular"}
        image_model = ImageModel(
            self.pixel_grid,
            self.psf_class,
            lens_light_model_class=self.lightModel,
            kwargs_numerics=kwargs_numerics,
        )
        from lenstronomy.ImSim.Numerics.grid import RegularGrid

        grid_class = image_model.ImageNumerics.grid_class
        assert isinstance(grid_class, RegularGrid)

        kwargs_numerics = {"compute_mode": "adaptive"}
        image_model = ImageModel(
            self.pixel_grid,
            self.psf_class,
            lens_light_model_class=self.lightModel,
            kwargs_numerics=kwargs_numerics,
        )
        from lenstronomy.ImSim.Numerics.grid import AdaptiveGrid

        grid_class = image_model.ImageNumerics.grid_class
        assert isinstance(grid_class, AdaptiveGrid)


def test_supersampling_simple():
    """

    :return:
    """
    from lenstronomy.Data.psf import PSF
    from lenstronomy.SimulationAPI.data_api import DataAPI

    detector_pixel_scale = 0.04
    numpix = 64
    supersampling_factor = 2
    # generate a Gaussian image

    x, y = util.make_grid(
        numPix=numpix * supersampling_factor,
        deltapix=detector_pixel_scale / supersampling_factor,
    )
    from lenstronomy.LightModel.Profiles.gaussian import Gaussian

    gaussian = Gaussian()
    image_1d = gaussian.function(x, y, amp=1, sigma=0.1)
    image = util.array2image(image_1d)

    # generate psf kernal supersampled
    kernel_super = kernel_util.kernel_gaussian(
        num_pix=21 * supersampling_factor + 1,
        delta_pix=detector_pixel_scale / supersampling_factor,
        fwhm=0.2,
    )

    psf_parameters = {
        "psf_type": "PIXEL",
        "kernel_point_source": kernel_super,
        "point_source_supersampling_factor": supersampling_factor,
    }
    kwargs_detector = {
        "pixel_scale": detector_pixel_scale,
        "ccd_gain": 2.5,
        "read_noise": 4.0,
        "magnitude_zero_point": 25.0,
        "exposure_time": 5400.0,
        "sky_brightness": 22,
        "num_exposures": 1,
        "background_noise": None,
    }
    kwargs_numerics = {
        "supersampling_factor": 2,
        "supersampling_convolution": True,
        "point_source_supersampling_factor": 2,
        "supersampling_kernel_size": 21,
    }
    psf_model = PSF(**psf_parameters)
    data_class = DataAPI(numpix=numpix, **kwargs_detector).data_class

    from lenstronomy.ImSim.Numerics.numerics_subframe import NumericsSubFrame

    image_numerics = NumericsSubFrame(
        pixel_grid=data_class, psf=psf_model, **kwargs_numerics
    )

    conv_class = image_numerics.convolution_class
    conv_flat = conv_class.convolution2d(image)
    print(np.shape(conv_flat), "shape of output")

    # psf_helper = lenstronomy_utils.PSFHelper(data_class, psf_model, kwargs_numerics)

    # Convolve with lenstronomy and with scipy
    # helper_image = psf_helper.psf_model(image)
    from scipy import signal

    scipy_image = signal.fftconvolve(image, kernel_super, mode="same")
    from lenstronomy.Util import image_util

    image_scipy_resized = image_util.re_size(scipy_image, supersampling_factor)
    image_unconvolved = image_util.re_size(image, supersampling_factor)

    # Compare the outputs

    # low res convolution as comparison
    kwargs_numerics_low_res = {
        "supersampling_factor": 2,
        "supersampling_convolution": False,
        "point_source_supersampling_factor": 2,
    }
    image_numerics_low_res = NumericsSubFrame(
        pixel_grid=data_class, psf=psf_model, **kwargs_numerics_low_res
    )
    conv_class_low_res = image_numerics_low_res.convolution_class
    conv_flat_low_res = conv_class_low_res.convolution2d(image_unconvolved)

    # import matplotlib.pyplot as plt
    # plt.matshow(image_scipy_resized - image_unconvolved)
    # plt.colorbar()
    # plt.show()

    # plt.matshow(image_scipy_resized - conv_flat)
    # plt.colorbar()
    # plt.show()

    # plt.matshow(image_scipy_resized - conv_flat_low_res)
    # plt.colorbar()
    # plt.show()

    np.testing.assert_almost_equal(conv_flat, image_scipy_resized)


class TestRaise(unittest.TestCase):
    def test_integer_in_supersampling_factor(self):
        from lenstronomy.Data.psf import PSF

        kwargs_psf = {"psf_type": "NONE"}
        psf_class = PSF(**kwargs_psf)

        from lenstronomy.ImSim.Numerics.numerics import Numerics

        with self.assertRaises(TypeError):
            Numerics(pixel_grid=None, psf=psf_class, supersampling_factor=1.0)


if __name__ == "__main__":
    pytest.main()



__author__ = "dgilman"

import numpy as np
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["CoreBurkert"]


class CoreBurkert(LensProfileBase):
    """Lensing properties of a modified Burkert profile with variable core size
    normalized by rho0, the central core density."""

    param_names = ["Rs", "alpha_Rs", "r_core", "center_x", "center_y"]
    lower_limit_default = {
        "Rs": 1,
        "alpha_Rs": 0,
        "r_core": 0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "Rs": 100,
        "alpha_Rs": 100,
        "r_core": 50,
        "center_x": 100,
        "center_y": 100,
    }

    def function(self, x, y, Rs, alpha_Rs, r_core, center_x=0, center_y=0):
        """

        :param x: angular position
        :param y: angular position
        :param Rs: angular turn over point
        :param alpha_Rs: deflection angle at Rs
        :param center_x: center of halo
        :param center_y: center of halo
        :return:
        """

        rho0 = self._alpha2rho0(alpha_Rs=alpha_Rs, Rs=Rs, r_core=r_core)

        if Rs < 0.0000001:
            Rs = 0.0000001
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        f_ = self.cBurkPot(R, Rs, rho0, r_core)
        return f_

    def derivatives(self, x, y, Rs, alpha_Rs, r_core, center_x=0, center_y=0):
        """Deflection angles :param x: x coordinate :param y: y coordinate :param Rs:
        scale radius :param alpha_Rs: deflection angle at Rs :param r_core: core radius
        :param center_x:

        :param center_y:
        :return:
        """

        rho0 = self._alpha2rho0(alpha_Rs=alpha_Rs, Rs=Rs, r_core=r_core)

        if Rs < 0.0000001:
            Rs = 0.0000001
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)

        dx, dy = self.coreBurkAlpha(R, Rs, rho0, r_core, x_, y_)

        return dx, dy

    def hessian(self, x, y, Rs, alpha_Rs, r_core, center_x=0, center_y=0):
        """
        :param x: x coordinate
        :param y: y coordinate
        :param Rs: scale radius
        :param alpha_Rs: deflection angle at Rs
        :param r_core: core radius
        :param center_x:
        :param center_y:
        :return:
        """

        if Rs < 0.0001:
            Rs = 0.0001
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)

        rho0 = self._alpha2rho0(alpha_Rs=alpha_Rs, Rs=Rs, r_core=r_core)

        kappa = self.density_2d(x_, y_, Rs, rho0, r_core)

        gamma1, gamma2 = self.cBurkGamma(R, Rs, rho0, r_core, x_, y_)
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_xy, f_xy, f_yy

    def mass_2d(self, R, Rs, rho0, r_core):
        """Analytic solution of the projection integral (convergence)

        :param R: projected distance
        :param Rs: scale radius
        :param rho0: central core density
        :param r_core: core radius
        """

        x = R * Rs**-1
        p = Rs * r_core**-1
        gx = self._G(x, p)

        m_2d = 2 * np.pi * rho0 * Rs**3 * gx

        return m_2d

    def coreBurkAlpha(self, R, Rs, rho0, r_core, ax_x, ax_y):
        """Deflection angle.

        :param R:
        :param Rs:
        :param rho0:
        :param r_core:
        :param ax_x:
        :param ax_y:
        :return:
        """
        x = R * Rs**-1
        p = Rs * r_core**-1

        gx = self._G(x, p)

        a = 2 * rho0 * Rs**2 * gx / x

        return a * ax_x / R, a * ax_y / R

    def density(self, R, Rs, rho0, r_core):
        """Three dimensional cored Burkert profile.

        :param R: radius of interest
        :type R: float/numpy array
        :param Rs: scale radius
        :type Rs: float
        :param rho0: characteristic density
        :type rho0: float
        :return: rho(R) density
        """

        M0 = 4 * np.pi * Rs**3 * rho0

        return (M0 / (4 * np.pi)) * ((r_core + R) * (Rs**2 + R**2)) ** -1

    def density_2d(self, x, y, Rs, rho0, r_core, center_x=0, center_y=0):
        """Projected two dimenstional core Burkert profile (kappa*Sigma_crit)

        :param x: x coordinate
        :param y: y coordinate
        :param Rs: scale radius
        :param rho0: central core density
        :param r_core: core radius
        """
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        x = R * Rs**-1
        p = Rs * r_core**-1
        Fx = self._F(x, p)

        return 2 * rho0 * Rs * Fx

    def mass_3d(self, R, Rs, rho0, r_core):
        """
        :param R: projected distance
        :param Rs: scale radius
        :param rho0: central core density
        :param r_core: core radius
        """

        Rs = float(Rs)
        b = r_core * Rs**-1
        c = R * Rs**-1

        M0 = 4 * np.pi * Rs**3 * rho0

        return (
            M0
            * (1 + b**2) ** -1
            * (
                0.5 * np.log(1 + c**2)
                + b**2 * np.log(c * b**-1 + 1)
                - b * np.arctan(c)
            )
        )

    def cBurkPot(self, R, Rs, rho0, r_core):
        """
        :param R: projected distance
        :param Rs: scale radius
        :param rho0: central core density
        :param r_core: core radius
        """
        x = R * Rs**-1
        p = Rs * r_core**-1
        hx = self._H(x, p)

        return 2 * rho0 * Rs**3 * hx

    def cBurkGamma(self, R, Rs, rho0, r_core, ax_x, ax_y):
        """

        :param R: projected distance
        :param Rs: scale radius
        :param rho0: central core density
        :param r_core: core radius
        :param ax_x: x coordinate
        :param ax_y: y coordinate
        :return:
        """
        c = 0.000001

        if isinstance(R, int) or isinstance(R, float):
            R = max(R, c)
        else:
            R[R <= c] = c

        x = R * Rs**-1
        p = Rs * r_core**-1

        gx = self._G(x, p)
        fx = self._F(x, p)

        m_x = 2 * rho0 * Rs**3 * gx
        kappa = 2 * rho0 * Rs * fx

        a = 2 * (m_x * R**-2 - kappa)

        return 0.5 * a * (ax_y**2 - ax_x**2) / R**2, -a * (ax_x * ax_y) / R**2

    @staticmethod
    def _u(x):
        return np.sqrt(1 + x**2)

    @staticmethod
    def _g(x, p):
        return np.sqrt(1 - x**2 * p**2)

    @staticmethod
    def _f(x, p):
        return np.sqrt(x**2 * p**2 - 1)

    def _H(self, x, p):
        prefactor = (p + p**3) ** -1 * p

        if isinstance(x, np.ndarray):
            inds0 = np.where(x * p == 1)
            inds1 = np.where(x * p < 1)
            inds2 = np.where(x * p > 1)
            func = np.ones_like(x)

            func[inds1] = (
                0.9058413472016891
                + (
                    -0.9640065632861909
                    + np.pi * self._u(x[inds1])
                    - 0.9058413472016892 * p
                )
                * p
                + 2
                * p**2
                * (self._u(x[inds1]) - 0.5 * np.arctanh(self._u(x[inds1]) ** -1))
                * np.arctanh(self._u(x[inds1]) ** -1)
                + 2
                * (self._g(x[inds1], p) - 0.5 * np.arctanh(self._g(x[inds1], p)))
                * np.arctanh(self._g(x[inds1], p))
                + (1 + p**2) * np.log(x[inds1]) ** 2
                - np.pi * p * np.log(1 + self._u(x[inds1]))
                + (0.3068528194400547 + 0.25 * np.log(p**2)) * np.log(p**2)
                + np.log(x[inds1])
                * (0.6137056388801094 + 0.6137056388801094 * p**2 + np.log(p**2))
            )

            func[inds2] = (
                0.9058413472016891
                + (
                    -0.9640065632861909
                    + np.pi * self._u(x[inds2])
                    - 0.9058413472016892 * p
                )
                * p
                + 2
                * p**2
                * (self._u(x[inds2]) - 0.5 * np.arctanh(self._u(x[inds2]) ** -1))
                * np.arctanh(self._u(x[inds2]) ** -1)
                + -2
                * (self._f(x[inds2], p) - 0.5 * np.arctan(self._f(x[inds2], p)))
                * np.arctan(self._f(x[inds2], p))
                + (1 + p**2) * np.log(x[inds2]) ** 2
                - np.pi * p * np.log(1 + self._u(x[inds2]))
                + (0.3068528194400547 + 0.25 * np.log(p**2)) * np.log(p**2)
                + np.log(x[inds2])
                * (0.6137056388801094 + 0.6137056388801094 * p**2 + np.log(p**2))
            )

            func[inds0] = (
                0.9058413472016891
                + (
                    -0.9640065632861909
                    + np.pi * self._u(x[inds0])
                    - 0.9058413472016892 * p
                )
                * p
                + 2
                * p**2
                * (self._u(x[inds0]) - 0.5 * np.arctanh(self._u(x[inds0]) ** -1))
                * np.arctanh(self._u(x[inds0]) ** -1)
                + (1 + p**2) * np.log(x[inds0]) ** 2
                - np.pi * p * np.log(1 + self._u(x[inds0]))
                + (0.3068528194400547 + 0.25 * np.log(p**2)) * np.log(p**2)
                + np.log(x[inds0])
                * (0.6137056388801094 + 0.6137056388801094 * p**2 + np.log(p**2))
            )

        else:
            if x * p < 1:
                func = (
                    0.9058413472016891
                    + (
                        -0.9640065632861909
                        + np.pi * self._u(x)
                        - 0.9058413472016892 * p
                    )
                    * p
                    + 2
                    * p**2
                    * (self._u(x) - 0.5 * np.arctanh(self._u(x) ** -1))
                    * np.arctanh(self._u(x) ** -1)
                    + 2
                    * (self._g(x, p) - 0.5 * np.arctanh(self._g(x, p)))
                    * np.arctanh(self._g(x, p))
                    + (1 + p**2) * np.log(x) ** 2
                    - np.pi * p * np.log(1 + self._u(x))
                    + (0.3068528194400547 + 0.25 * np.log(p**2)) * np.log(p**2)
                    + np.log(x)
                    * (
                        0.6137056388801094
                        + 0.6137056388801094 * p**2
                        + np.log(p**2)
                    )
                )
            elif x * p > 1:
                func = (
                    0.9058413472016891
                    + (
                        -0.9640065632861909
                        + np.pi * self._u(x)
                        - 0.9058413472016892 * p
                    )
                    * p
                    + 2
                    * p**2
                    * (self._u(x) - 0.5 * np.arctanh(self._u(x) ** -1))
                    * np.arctanh(self._u(x) ** -1)
                    + -2
                    * (self._f(x, p) - 0.5 * np.arctan(self._f(x, p)))
                    * np.arctan(self._f(x, p))
                    + (1 + p**2) * np.log(x) ** 2
                    - np.pi * p * np.log(1 + self._u(x))
                    + (0.3068528194400547 + 0.25 * np.log(p**2)) * np.log(p**2)
                    + np.log(x)
                    * (
                        0.6137056388801094
                        + 0.6137056388801094 * p**2
                        + np.log(p**2)
                    )
                )
            else:
                func = (
                    0.9058413472016891
                    + (
                        -0.9640065632861909
                        + np.pi * self._u(x)
                        - 0.9058413472016892 * p
                    )
                    * p
                    + 2
                    * p**2
                    * (self._u(x) - 0.5 * np.arctanh(self._u(x) ** -1))
                    * np.arctanh(self._u(x) ** -1)
                    + (1 + p**2) * np.log(x) ** 2
                    - np.pi * p * np.log(1 + self._u(x))
                    + (0.3068528194400547 + 0.25 * np.log(p**2)) * np.log(p**2)
                    + np.log(x)
                    * (
                        0.6137056388801094
                        + 0.6137056388801094 * p**2
                        + np.log(p**2)
                    )
                )

        return prefactor * func

    def _F(self, x, p):
        """Solution of the projection integal (kappa) arctanh / arctan function :param
        x: r/Rs :param p: r_core / Rs :return:"""
        prefactor = 0.5 * (1 + p**2) ** -1 * p

        if isinstance(x, np.ndarray):
            inds0 = np.where(x * p == 1)
            inds1 = np.where(x * p < 1)
            inds2 = np.where(x * p > 1)

            func = np.ones_like(x)

            func[inds0] = self._u(x[inds0]) ** -1 * (
                np.pi + 2 * p * np.arctanh(self._u(x[inds0]) ** -1)
            )

            func[inds1] = self._u(x[inds1]) ** -1 * (
                np.pi + 2 * p * np.arctanh(self._u(x[inds1]) ** -1)
            ) - (2 * p * self._g(x[inds1], p) ** -1 * np.arctanh(self._g(x[inds1], p)))

            func[inds2] = self._u(x[inds2]) ** -1 * (
                np.pi + 2 * p * np.arctanh(self._u(x[inds2]) ** -1)
            ) - (2 * p * self._f(x[inds2], p) ** -1 * np.arctan(self._f(x[inds2], p)))

            return prefactor * func

        else:
            if x * p == 1:
                func = self._u(x) ** -1 * (np.pi + 2 * p * np.arctanh(self._u(x) ** -1))
            elif x * p < 1:
                func = self._u(x) ** -1 * (
                    np.pi + 2 * p * np.arctanh(self._u(x) ** -1)
                ) - (2 * p * self._g(x, p) ** -1 * np.arctanh(self._g(x, p)))
            else:
                func = self._u(x) ** -1 * (
                    np.pi + 2 * p * np.arctanh(self._u(x) ** -1)
                ) - (2 * p * self._f(x, p) ** -1 * np.arctan(self._f(x, p)))

            return prefactor * func

    def _G(self, x, p):
        """
        analytic solution of the 2d projected mass integral
        integral: 2 * pi * x * kappa * dx
        :param x:
        :param p:
        :return:
        """

        prefactor = (p + p**3) ** -1 * p

        if isinstance(x, np.ndarray):
            inds0 = np.where(x * p == 1)
            inds1 = np.where(x * p < 1)
            inds2 = np.where(x * p > 1)

            func = np.ones_like(x)

            func[inds0] = (
                np.log(0.25 * x[inds0] ** 2 * p**2)
                + np.pi * p * (self._u(x[inds0]) - 1)
                + 2
                * p**2
                * (
                    self._u(x[inds0]) * np.arctanh(self._u(x[inds0]) ** -1)
                    + np.log(0.5 * x[inds0])
                )
            )

            func[inds1] = (
                np.log(0.25 * x[inds1] ** 2 * p**2)
                + np.pi * p * (self._u(x[inds1]) - 1)
                + 2
                * p**2
                * (
                    self._u(x[inds1]) * np.arctanh(self._u(x[inds1]) ** -1)
                    + np.log(0.5 * x[inds1])
                )
                + 2 * self._g(x[inds1], p) * np.arctanh(self._g(x[inds1], p))
            )

            func[inds2] = (
                np.log(0.25 * x[inds2] ** 2 * p**2)
                + np.pi * p * (self._u(x[inds2]) - 1)
                + 2
                * p**2
                * (
                    self._u(x[inds2]) * np.arctanh(self._u(x[inds2]) ** -1)
                    + np.log(0.5 * x[inds2])
                )
                - 2 * self._f(x[inds2], p) * np.arctan(self._f(x[inds2], p))
            )

        else:
            if x * p == 1:
                func = (
                    np.log(0.25 * x**2 * p**2)
                    + np.pi * p * (self._u(x) - 1)
                    + 2
                    * p**2
                    * (self._u(x) * np.arctanh(self._u(x) ** -1) + np.log(0.5 * x))
                )

            elif x * p < 1:
                func = (
                    np.log(0.25 * x**2 * p**2)
                    + np.pi * p * (self._u(x) - 1)
                    + 2
                    * p**2
                    * (self._u(x) * np.arctanh(self._u(x) ** -1) + np.log(0.5 * x))
                    + 2 * self._g(x, p) * np.arctanh(self._g(x, p))
                )

            else:
                func = (
                    np.log(0.25 * x**2 * p**2)
                    + np.pi * p * (self._u(x) - 1)
                    + 2
                    * p**2
                    * (self._u(x) * np.arctanh(self._u(x) ** -1) + np.log(0.5 * x))
                    - 2 * self._f(x, p) * np.arctan(self._f(x, p))
                )

        return func * prefactor

    def _alpha2rho0(self, alpha_Rs=None, Rs=None, r_core=None):
        p = Rs * r_core**-1

        gx = self._G(1, p)

        rho0 = alpha_Rs / (2 * Rs**2 * gx)

        return rho0

    def _rho2alpha(self, rho0=None, Rs=None, r_core=None):
        p = Rs / r_core
        gx = self._G(1, p)
        alpha = 2 * Rs**2 * gx * rho0

        return alpha



__author__ = "sibirrer"


import numpy as np
import scipy.special as special
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["SPP"]


class SPP(LensProfileBase):
    """Class for circular power-law mass distribution."""

    param_names = ["theta_E", "gamma", "center_x", "center_y"]
    lower_limit_default = {
        "theta_E": 0,
        "gamma": 1.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "theta_E": 100,
        "gamma": 2.5,
        "center_x": 100,
        "center_y": 100,
    }

    def function(self, x, y, theta_E, gamma, center_x=0, center_y=0):
        """
        :param x: set of x-coordinates
        :type x: array of size (n)
        :param y: set of y-coordinates
        :type y: array of size (n)
        :param theta_E: Einstein radius of lens
        :type theta_E: float.
        :param gamma: power law slope of mass profile
        :type gamma: <2 float
        :returns:  function
        :raises: AttributeError, KeyError
        """
        gamma = self._gamma_limit(gamma)

        x_ = x - center_x
        y_ = y - center_y
        E = theta_E / ((3.0 - gamma) / 2.0) ** (1.0 / (1.0 - gamma))
        # E = phi_E_spp
        eta = -gamma + 3

        p2 = x_**2 + y_**2
        s2 = 0.0  # softening
        return 2 * E**2 / eta**2 * ((p2 + s2) / E**2) ** (eta / 2)

    def derivatives(self, x, y, theta_E, gamma, center_x=0.0, center_y=0.0):
        gamma = self._gamma_limit(gamma)

        xt1 = x - center_x
        xt2 = y - center_y

        r2 = xt1 * xt1 + xt2 * xt2
        a = np.maximum(r2, 0.000001)
        r = np.sqrt(a)
        alpha = theta_E * (r2 / theta_E**2) ** (1 - gamma / 2.0)
        fac = alpha / r
        f_x = fac * xt1
        f_y = fac * xt2
        return f_x, f_y

    def hessian(self, x, y, theta_E, gamma, center_x=0.0, center_y=0.0):
        gamma = self._gamma_limit(gamma)
        xt1 = x - center_x
        xt2 = y - center_y
        E = theta_E / ((3.0 - gamma) / 2.0) ** (1.0 / (1.0 - gamma))
        # E = phi_E_spp
        eta = -gamma + 3.0

        P2 = xt1**2 + xt2**2
        if isinstance(P2, int) or isinstance(P2, float):
            a = max(0.000001, P2)
        else:
            a = np.empty_like(P2)
            p2 = P2[P2 > 0]  # in the SIS regime
            a[P2 == 0] = 0.000001
            a[P2 > 0] = p2

        kappa = (
            1.0
            / eta
            * (a / E**2) ** (eta / 2 - 1)
            * ((eta - 2) * (xt1**2 + xt2**2) / a + (1 + 1))
        )
        gamma1 = (
            1.0
            / eta
            * (a / E**2) ** (eta / 2 - 1)
            * ((eta / 2 - 1) * (2 * xt1**2 - 2 * xt2**2) / a)
        )
        gamma2 = (
            4 * xt1 * xt2 * (1.0 / 2 - 1 / eta) * (a / E**2) ** (eta / 2 - 2) / E**2
        )

        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_xy, f_xy, f_yy

    @staticmethod
    def rho2theta(rho0, gamma):
        """Converts 3d density into 2d projected density parameter.

        :param rho0:
        :param gamma:
        :return:
        """
        fac = (
            np.sqrt(np.pi)
            * special.gamma(1.0 / 2 * (-1 + gamma))
            / special.gamma(gamma / 2.0)
            * 2
            / (3 - gamma)
            * rho0
        )

        # fac = theta_E**(gamma - 1)
        theta_E = fac ** (1.0 / (gamma - 1))
        return theta_E

    @staticmethod
    def theta2rho(theta_E, gamma):
        """Converts projected density parameter (in units of deflection) into 3d density
        parameter.

        :param theta_E:
        :param gamma:
        :return:
        """
        fac1 = (
            np.sqrt(np.pi)
            * special.gamma(1.0 / 2 * (-1 + gamma))
            / special.gamma(gamma / 2.0)
            * 2
            / (3 - gamma)
        )
        fac2 = theta_E ** (gamma - 1)
        rho0 = fac2 / fac1
        return rho0

    @staticmethod
    def mass_3d(r, rho0, gamma):
        """Mass enclosed a 3d sphere or radius r.

        :param r:
        :param rho0:
        :param gamma:
        :return:
        """
        mass_3d = 4 * np.pi * rho0 / (-gamma + 3) * r ** (-gamma + 3)
        return mass_3d

    def mass_3d_lens(self, r, theta_E, gamma):
        """

        :param r:
        :param theta_E:
        :param gamma:
        :return:
        """
        rho0 = self.theta2rho(theta_E, gamma)
        return self.mass_3d(r, rho0, gamma)

    def mass_2d(self, r, rho0, gamma):
        """Mass enclosed projected 2d sphere of radius r.

        :param r:
        :param rho0:
        :param gamma:
        :return:
        """
        alpha = (
            np.sqrt(np.pi)
            * special.gamma(1.0 / 2 * (-1 + gamma))
            / special.gamma(gamma / 2.0)
            * r ** (2 - gamma)
            / (3 - gamma)
            * 2
            * rho0
        )
        mass_2d = alpha * r * np.pi
        return mass_2d

    def mass_2d_lens(self, r, theta_E, gamma):
        """

        :param r: projected radius
        :param theta_E: Einstein radius
        :param gamma: power-law slope
        :return: 2d projected radius enclosed
        """
        rho0 = self.theta2rho(theta_E, gamma)
        return self.mass_2d(r, rho0, gamma)

    def grav_pot(self, x, y, rho0, gamma, center_x=0, center_y=0):
        """Gravitational potential (modulo 4 pi G and rho0 in appropriate units)

        :param x:
        :param y:
        :param rho0:
        :param gamma:
        :param center_x:
        :param center_y:
        :return:
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        mass_3d = self.mass_3d(r, rho0, gamma)
        pot = mass_3d / r
        return pot

    @staticmethod
    def density(r, rho0, gamma):
        """Computes the density.

        :param r:
        :param rho0:
        :param gamma:
        :return:
        """
        rho = rho0 / r**gamma
        return rho

    def density_lens(self, r, theta_E, gamma):
        """Computes the density at 3d radius r given lens model parameterization.

        The integral in projected in units of angles (i.e. arc seconds) results in the
        convergence quantity.
        """
        rho0 = self.theta2rho(theta_E, gamma)
        return self.density(r, rho0, gamma)

    @staticmethod
    def density_2d(x, y, rho0, gamma, center_x=0, center_y=0):
        """Projected density.

        :param x:
        :param y:
        :param rho0:
        :param gamma:
        :param center_x:
        :param center_y:
        :return:
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        sigma = (
            np.sqrt(np.pi)
            * special.gamma(1.0 / 2 * (-1 + gamma))
            / special.gamma(gamma / 2.0)
            * r ** (1 - gamma)
            * rho0
        )
        return sigma

    @staticmethod
    def _gamma_limit(gamma):
        """Limits the power-law slope to certain bounds.

        :param gamma: power-law slope
        :return: bounded power-law slopte
        """
        return gamma



from lenstronomy.GalKin.galkin import Galkin
from lenstronomy.GalKin.galkin_shells import GalkinShells
import numpy as np
import numpy.testing as npt
import pytest
import unittest


class TestGalkinShells(object):
    def test_vel_disp(self):
        # light profile
        light_profile_list = ["HERNQUIST"]
        Rs = 0.5
        kwargs_light = [
            {"Rs": Rs, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        # 0.551 *
        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 2.0
        kwargs_profile = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 2.0
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        # aperture as slit
        aperture_type = "IFU_shells"
        r_bins = np.linspace(start=0, stop=2, num=5)
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "r_bins": r_bins,
            "center_ra": 0,
            "center_dec": 0,
        }

        psf_fwhm = 0.7  # Gaussian FWHM psf
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_numerics_log = {
            "interpol_grid_num": 1000,
            "log_integration": True,
            "max_integrate": 10,
            "min_integrate": 0.001,
            "lum_weight_int_method": True,
        }

        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        galkin = Galkin(
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics_log,
        )
        galkin_shells = GalkinShells(
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics_log,
        )

        vel_disp_bins = galkin_shells.dispersion_map(
            kwargs_mass=kwargs_profile,
            kwargs_light=kwargs_light,
            kwargs_anisotropy=kwargs_anisotropy,
        )
        disp_map = galkin.dispersion_map(
            kwargs_mass=kwargs_profile,
            kwargs_light=kwargs_light,
            kwargs_anisotropy=kwargs_anisotropy,
            num_kin_sampling=1000,
            num_psf_sampling=100,
        )
        npt.assert_almost_equal(vel_disp_bins / disp_map, 1, decimal=2)


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            kwargs_model = {
                "mass_profile_list": ["SPP"],
                "light_profile_list": ["HERNQUIST"],
                "anisotropy_model": "const",
            }
            kwargs_aperture = {
                "center_ra": 0,
                "width": 1,
                "length": 1,
                "angle": 0,
                "center_dec": 0,
                "aperture_type": "slit",
            }
            kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
            kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": 1}
            GalkinShells(kwargs_model, kwargs_aperture, kwargs_psf, kwargs_cosmo)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"


import numpy as np
import pytest
import numpy.testing as npt
import lenstronomy.Util.param_util as param_util


class TestEPL_numba(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        from lenstronomy.LensModel.Profiles.epl import EPL

        self.EPL = EPL()
        from lenstronomy.LensModel.Profiles.epl_numba import EPL_numba

        self.EPL_numba = EPL_numba()

    def test_function(self):
        phi_E = 1.0
        gamma = 2.0
        q = 0.999
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        x = np.array([1.0, 2])
        y = np.array([2, 0])
        values = self.EPL.function(x, y, phi_E, gamma, e1, e2)
        values_nb = self.EPL_numba.function(x, y, phi_E, gamma, e1, e2)
        delta_f = values[0] - values[1]
        delta_f_nb = values_nb[0] - values_nb[1]
        npt.assert_almost_equal(delta_f, delta_f_nb, decimal=10)

        q = 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        x = np.array([1.0, 2])
        y = np.array([2, 0])
        values = self.EPL.function(x, y, phi_E, gamma, e1, e2)
        values_nb = self.EPL_numba.function(x, y, phi_E, gamma, e1, e2)
        delta_f = values[0] - values[1]
        delta_f_nb = values_nb[0] - values_nb[1]
        npt.assert_almost_equal(delta_f, delta_f_nb, decimal=10)

        q = 0.4
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        x = np.array([1.0, 2])
        y = np.array([2, 0])
        values = self.EPL.function(x, y, phi_E, gamma, e1, e2)
        values_nb = self.EPL_numba.function(x, y, phi_E, gamma, e1, e2)
        delta_f = values[0] - values[1]
        delta_f_nb = values_nb[0] - values_nb[1]
        npt.assert_almost_equal(delta_f, delta_f_nb, decimal=10)

    def test_derivatives(self):
        x = np.array([1])
        y = np.array([2])
        phi_E = 1.0
        gamma = 1.8
        q = 1.0
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_x, f_y = self.EPL.derivatives(x, y, phi_E, gamma, e1, e2)
        f_x_nb, f_y_nb = self.EPL_numba.derivatives(x, y, phi_E, gamma, e1, e2)
        npt.assert_almost_equal(f_x, f_x_nb, decimal=10)
        npt.assert_almost_equal(f_y, f_y_nb, decimal=10)

        q = 0.7
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_x, f_y = self.EPL.derivatives(x, y, phi_E, gamma, e1, e2)
        f_x_nb, f_y_nb = self.EPL_numba.derivatives(x, y, phi_E, gamma, e1, e2)
        npt.assert_almost_equal(f_x, f_x_nb, decimal=10)
        npt.assert_almost_equal(f_y, f_y_nb, decimal=10)

    def test_hessian(self):
        x = np.array([1.0])
        y = np.array([2.0])
        phi_E = 1.0
        gamma = 2.2
        q = 0.9
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_xx, f_xy, f_yx, f_yy = self.EPL.hessian(x, y, phi_E, gamma, e1, e2)
        f_xx_nb, f_xy_nb, f_yx_nb, f_yy_nb = self.EPL_numba.hessian(
            x, y, phi_E, gamma, e1, e2
        )
        npt.assert_almost_equal(f_xx, f_xx_nb, decimal=10)
        npt.assert_almost_equal(f_yy, f_yy_nb, decimal=10)
        npt.assert_almost_equal(f_xy, f_xy_nb, decimal=10)

    def test_regularization(self):
        phi_E = 1.0
        gamma = 2.0
        q = 1.0
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)

        x = 0.0
        y = 0.0
        f_x, f_y = self.EPL_numba.derivatives(x, y, phi_E, gamma, e1, e2)
        npt.assert_almost_equal(f_x, 0.0)
        npt.assert_almost_equal(f_y, 0.0)

        x = 0.0
        y = 0.0
        f_x, f_y = self.EPL.derivatives(x, y, phi_E, gamma, e1, e2)
        npt.assert_almost_equal(f_x, 0.0)
        npt.assert_almost_equal(f_y, 0.0)

        x = 0.0
        y = 0.0
        f_x, f_y = self.EPL.derivatives(x, y, phi_E, gamma + 0.1, e1, e2)
        npt.assert_almost_equal(f_x, 0.0)
        npt.assert_almost_equal(f_y, 0.0)

        x = 0.0
        y = 0.0
        f = self.EPL_numba.function(x, y, phi_E, gamma, e1, e2)
        npt.assert_almost_equal(f, 0.0)

        x = 0.0
        y = 0.0
        f_xx, f_xy, f_yx, f_yy = self.EPL_numba.hessian(x, y, phi_E, gamma, e1, e2)
        npt.assert_almost_equal(f_xx, 1e10, decimal=10)
        npt.assert_almost_equal(f_yy, 0, decimal=10)
        npt.assert_almost_equal(
            f_xy, 0, decimal=5
        )  # floating point cancellation, so less precise
        # Magnification:
        npt.assert_almost_equal(
            1 / ((1 - f_xx) * (1 - f_yy) - f_xy**2), 0.0, decimal=10
        )


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import pytest
import numpy.testing as npt
import numpy as np
import scipy.integrate as integrate


class TestNumerics(object):
    """Tests the second derivatives of various lens models."""

    def setup_method(self):
        pass

    def assert_integrals(self, Model, kwargs):
        lightModel = Model()
        r = 2.0
        out = integrate.quad(
            lambda x: 2 * lightModel.light_3d(np.sqrt(x**2 + r**2), **kwargs),
            0,
            100,
        )
        light_2d_num = out[0]
        light_2d = lightModel.function(r, 0, **kwargs)
        npt.assert_almost_equal(light_2d_num / light_2d, 1.0, decimal=1)

    def test_PJaffe(self):
        kwargs = {"amp": 1.0, "Ra": 0.2, "Rs": 2.0}
        from lenstronomy.LightModel.Profiles.p_jaffe import PJaffe as Model

        self.assert_integrals(Model, kwargs)

    def test_hernquist(self):
        kwargs = {"amp": 1.0, "Rs": 5.0}
        from lenstronomy.LightModel.Profiles.hernquist import Hernquist as Model

        self.assert_integrals(Model, kwargs)

    def test_gaussian(self):
        from lenstronomy.LightModel.Profiles.gaussian import Gaussian as Model

        kwargs = {"amp": 1.0 / 4.0, "sigma": 2.0}
        self.assert_integrals(Model, kwargs)

    def test_power_law(self):
        from lenstronomy.LightModel.Profiles.power_law import PowerLaw as Model

        kwargs = {"amp": 2, "gamma": 2, "e1": 0, "e2": 0}
        self.assert_integrals(Model, kwargs)

    def test_nie(self):
        from lenstronomy.LightModel.Profiles.nie import NIE as Model

        kwargs = {"amp": 2, "s_scale": 0.001, "e1": 0, "e2": 0}
        self.assert_integrals(Model, kwargs)
        kwargs = {"amp": 2, "s_scale": 1.0, "e1": 0, "e2": 0}
        self.assert_integrals(Model, kwargs)

    def test_chameleon(self):
        from lenstronomy.LightModel.Profiles.chameleon import Chameleon as Model

        kwargs = {"amp": 2, "w_c": 1, "w_t": 2, "e1": 0, "e2": 0}
        self.assert_integrals(Model, kwargs)

        from lenstronomy.LightModel.Profiles.chameleon import DoubleChameleon as Model

        kwargs = {
            "amp": 2,
            "ratio": 0.4,
            "w_c1": 1,
            "w_t1": 2,
            "e11": 0,
            "e21": 0,
            "w_c2": 2,
            "w_t2": 3,
            "e12": 0,
            "e22": 0,
        }
        self.assert_integrals(Model, kwargs)

        from lenstronomy.LightModel.Profiles.chameleon import TripleChameleon as Model

        kwargs = {
            "amp": 2,
            "ratio12": 0.4,
            "ratio13": 2,
            "w_c1": 1,
            "w_t1": 2,
            "e11": 0,
            "e21": 0,
            "w_c2": 2,
            "w_t2": 3,
            "e12": 0,
            "e22": 0,
            "w_c3": 0.2,
            "w_t3": 1,
            "e13": 0,
            "e23": 0,
        }
        self.assert_integrals(Model, kwargs)


if __name__ == "__main__":
    pytest.main("-k TestLensModel")



from lenstronomy.GalKin.cosmo import Cosmo

import pytest
import unittest


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            Cosmo(d_d=-1, d_s=1, d_ds=1)


if __name__ == "__main__":
    pytest.main()



__author__ = "lynevdv"

import numpy as np
import cmath as c
from lenstronomy.Util import param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["ElliSLICE"]


class ElliSLICE(LensProfileBase):
    """
    This class computes the lensing quantities for an elliptical slice of constant density.
    Based on Schramm 1994 https://ui.adsabs.harvard.edu/abs/1994A%26A...284...44S/abstract

    Computes the lensing quantities of an elliptical slice with semi major axis 'a' and
    semi minor axis 'b', centered on 'center_x' and 'center_y', oriented with an angle 'psi'
    in radian, and with constant surface mass density 'sigma_0'.
    In other words, this lens model is characterized by the surface mass density :

    ..math::
        \\kappa(x,y) = \\left{
            \\begin{array}{ll}
                \\sigma_0  & \\mbox{if } \\frac{x_{rot}^2}{a^2} + \\frac{y_{rot}^2}{b^2} \\leq 1 \\\
                0 & \\mbox{else}

            \\end{array}

        \\right}.

    with

    ..math::
        x_{rot} = x_c \\cos \\psi + y_c \\sin \\psi  \\
        y_{rot} = - x_c \\sin \\psi + y_c \\cos \\psi  \\
        x_c = x - center_x  \\
        y_c = y - center_y

    """

    param_names = ["a", "b", "psi", "sigma_0", "center_x", "center_y"]
    lower_limit_default = {
        "a": 0.0,
        "b": 0.0,
        "psi": -90.0 / 180.0 * np.pi,
        "center_x": -100.0,
        "center_y": -100.0,
    }
    upper_limit_default = {
        "a": 100.0,
        "b": 100.0,
        "psi": 90.0 / 180.0 * np.pi,
        "center_x": 100.0,
        "center_y": 100.0,
    }

    def function(self, x, y, a, b, psi, sigma_0, center_x=0.0, center_y=0.0):
        """Lensing potential.

        :param a: float, semi-major axis, must be positive
        :param b: float, semi-minor axis, must be positive
        :param psi: float, orientation in radian
        :param sigma_0: float, surface mass density, must be positive
        :param center_x: float, center on the x axis
        :param center_y: float, center on the y axis
        """
        kwargs_slice = {
            "center_x": center_x,
            "center_y": center_y,
            "a": a,
            "b": b,
            "psi": psi,
            "sigma_0": sigma_0,
        }
        x_ = x - center_x
        y_ = y - center_y
        x_rot = x_ * np.cos(psi) + y_ * np.sin(psi)
        y_rot = -x_ * np.sin(psi) + y_ * np.cos(psi)
        try:
            len(x_)
        except:
            if (x_rot**2 / a**2) + (y_rot**2 / b**2) <= 1:
                return self.pot_in(x_, y_, kwargs_slice)
            else:
                return self.pot_ext(x_, y_, kwargs_slice)
        else:
            f = np.array(
                [
                    self.pot_in(x_[i], y_[i], kwargs_slice)
                    if (x_rot[i] ** 2 / a**2) + (y_rot[i] ** 2 / b**2) <= 1
                    else self.pot_ext(x_[i], y_[i], kwargs_slice)
                    for i in range(len(x_))
                ]
            )
            return f

    def derivatives(self, x, y, a, b, psi, sigma_0, center_x=0.0, center_y=0.0):
        """Lensing deflection angle.

        :param a: float, semi-major axis, must be positive
        :param b: float, semi-minor axis, must be positive
        :param psi: float, orientation in radian
        :param sigma_0: float, surface mass density, must be positive
        :param center_x: float, center on the x axis
        :param center_y: float, center on the y axis
        """
        kwargs_slice = {
            "center_x": center_x,
            "center_y": center_y,
            "a": a,
            "b": b,
            "psi": psi,
            "sigma_0": sigma_0,
        }
        x_ = x - center_x
        y_ = y - center_y
        x_rot = x_ * np.cos(psi) + y_ * np.sin(psi)
        y_rot = -x_ * np.sin(psi) + y_ * np.cos(psi)
        try:
            len(x_)
        except:
            if (x_rot**2 / a**2) + (y_rot**2 / b**2) <= 1:
                return self.alpha_in(x_, y_, kwargs_slice)
            else:
                return self.alpha_ext(x_, y_, kwargs_slice)
        else:
            defl = np.array(
                [
                    self.alpha_in(x_[i], y_[i], kwargs_slice)
                    if (x_rot[i] ** 2 / a**2) + (y_rot[i] ** 2 / b**2) <= 1
                    else self.alpha_ext(x_[i], y_[i], kwargs_slice)
                    for i in range(len(x_))
                ]
            )
            return defl[:, 0], defl[:, 1]

    def hessian(self, x, y, a, b, psi, sigma_0, center_x=0.0, center_y=0.0):
        """Lensing second derivatives.

        :param a: float, semi-major axis, must be positive
        :param b: float, semi-minor axis, must be positive
        :param psi: float, orientation in radian
        :param sigma_0: float, surface mass density, must be positive
        :param center_x: float, center on the x axis
        :param center_y: float, center on the y axis
        """

        diff = 0.000000001
        alpha_ra, alpha_dec = self.derivatives(
            x, y, a, b, psi, sigma_0, center_x, center_y
        )
        alpha_ra_dx, alpha_dec_dx = self.derivatives(
            x + diff, y, a, b, psi, sigma_0, center_x, center_y
        )
        alpha_ra_dy, alpha_dec_dy = self.derivatives(
            x, y + diff, a, b, psi, sigma_0, center_x, center_y
        )

        f_xx = (alpha_ra_dx - alpha_ra) / diff
        f_xy = (alpha_ra_dy - alpha_ra) / diff
        f_yx = (alpha_dec_dx - alpha_dec) / diff
        f_yy = (alpha_dec_dy - alpha_dec) / diff
        return f_xx, f_xy, f_yx, f_yy

    @staticmethod
    def sign(z):
        """Sign function.

        :param z: complex
        """
        x = z.real
        y = z.imag
        if x > 0 or (x == 0 and y >= 0):
            return 1
        else:
            return -1

    def alpha_in(self, x, y, kwargs_slice):
        """Deflection angle for (x,y) inside the elliptical slice.

        :param kwargs_slice: dict, dictionary with the slice definition
            (a,b,psi,sigma_0)
        """
        z = complex(x, y)
        zb = z.conjugate()
        psi = kwargs_slice["psi"]
        e = (kwargs_slice["a"] - kwargs_slice["b"]) / (
            kwargs_slice["a"] + kwargs_slice["b"]
        )
        sig_0 = kwargs_slice["sigma_0"]
        e2ipsi = c.exp(2j * psi)
        I_in = (z - e * zb * e2ipsi) * sig_0
        return I_in.real, I_in.imag

    def alpha_ext(self, x, y, kwargs_slice):
        """Deflection angle for (x,y) outside the elliptical slice.

        :param kwargs_slice: dict, dictionary with the slice definition
            (a,b,psi,sigma_0)
        """
        z = complex(x, y)
        r, phi = param_util.cart2polar(x, y)
        zb = z.conjugate()
        psi = kwargs_slice["psi"]
        a = kwargs_slice["a"]
        b = kwargs_slice["b"]
        f2 = a**2 - b**2
        sig_0 = kwargs_slice["sigma_0"]
        median_op = False
        # when (x,y) is on one of the ellipse axis, there might be an issue when calculating the square root of
        # zb ** 2 * e2ipsi - f2. When the argument has an imaginary part ==0, having 0. or -0. may return different
        # answers. Therefore, for points (x,y) close to one axis, we take 3 points (one is x,y ; another one is a delta
        # away from this position, perpendicularly to the axis ; another one is at -delta perpendicularly away from
        # x,y). We calculate the function for each point and take the median. This avoids any singularity for points
        # along the axis but it slows down the function.
        if (
            np.abs(np.sin(phi - psi)) <= 10**-10
            or np.abs(np.sin(phi - psi - np.pi / 2.0)) <= 10**-10
        ):  # very close to one of the ellipse axis
            median_op = True
        e2ipsi = c.exp(2j * psi)
        eipsi = c.exp(1j * psi)
        if median_op is True:
            eps = 10**-10
            z_minus_eps = complex(r * np.cos(phi - eps), r * np.sin(phi - eps))
            zb_minus_eps = z_minus_eps.conjugate()
            z_plus_eps = complex(r * np.cos(phi + eps), r * np.sin(phi + eps))
            zb_plus_eps = z_plus_eps.conjugate()
            I_out_minus = (
                2
                * a
                * b
                / f2
                * (
                    zb_minus_eps * e2ipsi
                    - eipsi
                    * self.sign(zb_minus_eps * eipsi)
                    * c.sqrt(zb_minus_eps**2 * e2ipsi - f2)
                )
                * sig_0
            )
            I_out_plus = (
                2
                * a
                * b
                / f2
                * (
                    zb_plus_eps * e2ipsi
                    - eipsi
                    * self.sign(zb_plus_eps * eipsi)
                    * c.sqrt(zb_plus_eps**2 * e2ipsi - f2)
                )
                * sig_0
            )
            I_out_mid = (
                2
                * a
                * b
                / f2
                * (
                    zb * e2ipsi
                    - eipsi * self.sign(zb * eipsi) * c.sqrt(zb**2 * e2ipsi - f2)
                )
                * sig_0
            )
            I_out_real = np.median([I_out_minus.real, I_out_plus.real, I_out_mid.real])
            I_out_imag = np.median([I_out_minus.imag, I_out_plus.imag, I_out_mid.imag])
        else:
            I_out = (
                2
                * a
                * b
                / f2
                * (
                    zb * e2ipsi
                    - eipsi * self.sign(zb * eipsi) * c.sqrt(zb**2 * e2ipsi - f2)
                )
                * sig_0
            )
            I_out_real = I_out.real
            I_out_imag = I_out.imag

        # buf = zb ** 2 * e2ipsi - f2  ##problem with 0. and -0. giving different answers
        # if buf.real == 0:
        #     buf = complex(0, buf.imag)
        # if buf.imag == 0:
        #     buf = complex(buf.real, 0)

        return I_out_real, I_out_imag

    @staticmethod
    def pot_in(x, y, kwargs_slice):
        """Lensing potential for (x,y) inside the elliptical slice.

        :param kwargs_slice: dict, dictionary with the slice definition
            (a,b,psi,sigma_0)
        """
        psi = kwargs_slice["psi"]
        a = kwargs_slice["a"]
        b = kwargs_slice["b"]
        sig_0 = kwargs_slice["sigma_0"]
        e = (a - b) / (a + b)
        rE = (a + b) / 2.0
        pot_in = (
            0.5
            * (
                (1 - e) * (x * np.cos(psi) + y * np.sin(psi)) ** 2
                + (1 + e) * (y * np.cos(psi) - x * np.sin(psi)) ** 2
            )
            * sig_0
        )
        cst = sig_0 * rE**2 * (1 - e**2) * np.log(rE)
        return pot_in + cst

    def pot_ext(self, x, y, kwargs_slice):
        """Lensing potential for (x,y) outside the elliptical slice.

        :param kwargs_slice: dict, dictionary with the slice definition
            (a,b,psi,sigma_0)
        """
        z = complex(x, y)
        # zb = z.conjugate()
        psi = kwargs_slice["psi"]
        a = kwargs_slice["a"]
        b = kwargs_slice["b"]
        sig_0 = kwargs_slice["sigma_0"]
        r, phi = param_util.cart2polar(x, y)
        median_op = False
        # when (x,y) is on one of the ellipse axis, there might be an issue when calculating the square root of
        # z ** 2 * em2ipsi - f2. When the argument has an imaginary part ==0, having 0. or -0. may return different
        # answers. Therefore, for points (x,y) close to one axis, we take 3 points (one is x,y ; another one is a delta
        # away from this position, perpendicularly to the axis ; another one is at -delta perpendicularly away from
        # x,y). We calculate the function for each point and take the median. This avoids any singularity for points
        # along the axis but it slows down the function.
        if (
            np.abs(np.sin(phi - psi)) <= 10**-10
            or np.abs(np.sin(phi - psi - np.pi / 2.0)) <= 10**-10
        ):  # very close to one of the ellipse axis
            median_op = True
        e = (a - b) / (a + b)
        f2 = a**2 - b**2
        emipsi = c.exp(-1j * psi)
        em2ipsi = c.exp(-2j * psi)
        if median_op is True:
            eps = 10**-10
            z_minus_eps = complex(r * np.cos(phi - eps), r * np.sin(phi - eps))
            z_plus_eps = complex(r * np.cos(phi + eps), r * np.sin(phi + eps))

            pot_ext_minus = (
                (1 - e**2)
                / (4 * e)
                * (
                    f2
                    * c.log(
                        (
                            self.sign(z_minus_eps * emipsi) * z_minus_eps * emipsi
                            + c.sqrt(z_minus_eps**2 * em2ipsi - f2)
                        )
                        / 2.0
                    )
                    - self.sign(z_minus_eps * emipsi)
                    * z_minus_eps
                    * emipsi
                    * c.sqrt(z_minus_eps**2 * em2ipsi - f2)
                    + z_minus_eps**2 * em2ipsi
                )
                * sig_0
            )
            pot_ext_plus = (
                (1 - e**2)
                / (4 * e)
                * (
                    f2
                    * c.log(
                        (
                            self.sign(z_plus_eps * emipsi) * z_plus_eps * emipsi
                            + c.sqrt(z_plus_eps**2 * em2ipsi - f2)
                        )
                        / 2.0
                    )
                    - self.sign(z_plus_eps * emipsi)
                    * z_plus_eps
                    * emipsi
                    * c.sqrt(z_plus_eps**2 * em2ipsi - f2)
                    + z_plus_eps**2 * em2ipsi
                )
                * sig_0
            )
            pot_ext_mid = (
                (1 - e**2)
                / (4 * e)
                * (
                    f2
                    * c.log(
                        (
                            self.sign(z * emipsi) * z * emipsi
                            + c.sqrt(z**2 * em2ipsi - f2)
                        )
                        / 2.0
                    )
                    - self.sign(z * emipsi) * z * emipsi * c.sqrt(z**2 * em2ipsi - f2)
                    + z**2 * em2ipsi
                )
                * sig_0
            )
            pot_ext = np.median(
                [pot_ext_minus.real, pot_ext_plus.real, pot_ext_mid.real]
            )
        else:
            pot_ext = (
                (1 - e**2)
                / (4 * e)
                * (
                    f2
                    * c.log(
                        (
                            self.sign(z * emipsi) * z * emipsi
                            + c.sqrt(z**2 * em2ipsi - f2)
                        )
                        / 2.0
                    )
                    - self.sign(z * emipsi) * z * emipsi * c.sqrt(z**2 * em2ipsi - f2)
                    + z**2 * em2ipsi
                )
                * sig_0
            ).real
        return pot_ext



import pytest
import numpy as np
import numpy.testing as npt
from lenstronomy.LensModel.QuadOptimizer.param_manager import (
    PowerLawFixedShear,
    PowerLawFixedShearMultipole,
    PowerLawFreeShear,
    PowerLawFreeShearMultipole,
)


class TestParamClasses(object):
    def setup_method(self):
        self.zlens, self.zsource = 0.5, 1.5
        epl_kwargs = {
            "theta_E": 1.0,
            "center_x": 0.0,
            "center_y": 0.0,
            "e1": 0.2,
            "e2": 0.1,
            "gamma": 2.05,
        }
        shear_kwargs = {"gamma1": 0.05, "gamma2": -0.04}
        kwargs_macro = [epl_kwargs, shear_kwargs]

        self.x_image = np.array([0.65043538, -0.31109505, 0.78906059, -0.86222271])
        self.y_image = np.array([-0.89067493, 0.94851787, 0.52882605, -0.25403778])

        halo_list = ["SIS", "SIS", "SIS"]
        halo_z = [self.zlens - 0.1, self.zlens, self.zlens + 0.4]
        halo_kwargs = [
            {"theta_E": 0.1, "center_x": 0.3, "center_y": -0.9},
            {"theta_E": 0.15, "center_x": 1.3, "center_y": -0.5},
            {"theta_E": 0.06, "center_x": -0.4, "center_y": -0.4},
        ]

        self.kwargs_epl = kwargs_macro + halo_kwargs
        self.zlist_epl = [self.zlens, self.zlens] + halo_z
        self.lens_model_list_epl = ["EPL", "SHEAR"] + halo_list

        kwargs_multi = [
            {"m": 4, "a_m": -0.04, "phi_m": -0.2, "center_x": 0.1, "center_y": -0.1}
        ]
        self.kwargs_multipole = kwargs_macro + kwargs_multi + halo_kwargs
        self.zlist_multipole = [self.zlens, self.zlens, self.zlens] + halo_z
        self.lens_model_list_multipole = ["EPL", "SHEAR"] + ["MULTIPOLE"] + halo_list

    def test_param_penalty(self):
        param_class = PowerLawFreeShear(self.kwargs_epl)
        args = param_class.kwargs_to_args(self.kwargs_epl)
        param_penalty = param_class.param_chi_square_penalty(args)
        npt.assert_almost_equal(0, param_penalty)

    def test_plaw_free_shear(self):
        param_class = PowerLawFreeShear(self.kwargs_epl)
        npt.assert_(param_class.to_vary_index == 2)
        kwargs_in = [
            {
                "theta_E": 1.0,
                "center_x": 0.0,
                "center_y": 0.3,
                "e1": 0.25,
                "e2": 0.1,
                "gamma": 2.05,
            },
            {"gamma1": 0.05, "gamma2": -0.01},
            {"theta_E": -0.3, "center_x": 0.0, "center_y": 0.04},
        ]
        args_epl = param_class.kwargs_to_args(kwargs_in)
        npt.assert_almost_equal(args_epl, [1, 0, 0.3, 0.25, 0.1, 0.05, -0.01])
        kwargs_out = param_class.args_to_kwargs(args_epl)
        npt.assert_almost_equal(kwargs_out[0]["gamma"], 2.05)
        for key in kwargs_out[-1].keys():
            npt.assert_almost_equal(kwargs_out[-1][key], self.kwargs_epl[-1][key])

    def test_plaw_fixed_shear(self):
        param_class = PowerLawFixedShear(self.kwargs_epl, 0.12)
        npt.assert_(param_class.to_vary_index == 2)
        kwargs_in = [
            {
                "theta_E": 1.0,
                "center_x": 0.0,
                "center_y": 0.3,
                "e1": 0.25,
                "e2": 0.1,
                "gamma": 2.05,
            },
            {"gamma1": 0.05, "gamma2": -0.01},
            {"theta_E": -0.3, "center_x": 0.0, "center_y": 0.04},
        ]
        args_epl = param_class.kwargs_to_args(kwargs_in)
        npt.assert_almost_equal(args_epl[0:5], [1, 0, 0.3, 0.25, 0.1])
        kwargs_out = param_class.args_to_kwargs(args_epl)
        npt.assert_almost_equal(kwargs_out[0]["gamma"], 2.05)
        npt.assert_almost_equal(
            kwargs_out[1]["gamma1"] ** 2 + kwargs_out[1]["gamma2"] ** 2, 0.12**2
        )
        for key in kwargs_out[-1].keys():
            npt.assert_almost_equal(kwargs_out[-1][key], self.kwargs_epl[-1][key])

    def test_plawboxydisky_fixed_shear(self):
        param_class = PowerLawFixedShearMultipole(self.kwargs_multipole, 0.12)
        npt.assert_(param_class.to_vary_index == 3)
        kwargs_in = [
            {
                "theta_E": 1.0,
                "center_x": 0.0,
                "center_y": 0.3,
                "e1": 0.25,
                "e2": 0.1,
                "gamma": 2.05,
            },
            {"gamma1": 0.05, "gamma2": -0.01},
            {"theta_E": -0.3, "center_x": 0.0, "center_y": 0.04},
        ]
        args_epl = param_class.kwargs_to_args(kwargs_in)
        npt.assert_almost_equal(args_epl[0:5], [1, 0, 0.3, 0.25, 0.1])
        kwargs_out = param_class.args_to_kwargs(args_epl)
        npt.assert_almost_equal(kwargs_out[0]["gamma"], 2.05)
        npt.assert_almost_equal(
            kwargs_out[1]["gamma1"] ** 2 + kwargs_out[1]["gamma2"] ** 2, 0.12**2
        )
        for key in kwargs_out[-1].keys():
            npt.assert_almost_equal(kwargs_out[-1][key], self.kwargs_multipole[-1][key])

        for key in kwargs_out[2].keys():
            npt.assert_almost_equal(kwargs_out[2][key], self.kwargs_multipole[2][key])

    def test_plawboxydisky_fixed_shear(self):
        param_class = PowerLawFreeShearMultipole(self.kwargs_multipole)
        npt.assert_(param_class.to_vary_index == 3)
        kwargs_in = [
            {
                "theta_E": 1.0,
                "center_x": 0.0,
                "center_y": 0.3,
                "e1": 0.25,
                "e2": 0.1,
                "gamma": 2.05,
            },
            {"gamma1": 0.05, "gamma2": -0.01},
            {"theta_E": -0.3, "center_x": 0.0, "center_y": 0.04},
        ]
        args_epl = param_class.kwargs_to_args(kwargs_in)
        npt.assert_almost_equal(args_epl, [1, 0, 0.3, 0.25, 0.1, 0.05, -0.01])
        kwargs_out = param_class.args_to_kwargs(args_epl)
        npt.assert_almost_equal(kwargs_out[0]["gamma"], 2.05)
        for key in kwargs_out[-1].keys():
            npt.assert_almost_equal(kwargs_out[-1][key], self.kwargs_multipole[-1][key])

        for key in kwargs_out[2].keys():
            npt.assert_almost_equal(kwargs_out[2][key], self.kwargs_multipole[2][key])

    def test_bounds(self):
        param_names = ["theta_E", "center_x", "center_y", "e1", "e2"]
        args = [self.kwargs_epl[0][param_name] for param_name in param_names]
        param_names = ["gamma1", "gamma2"]
        args += [self.kwargs_epl[1][param_name] for param_name in param_names]

        shift = np.array([0.1, 0.05, 0.05, 0.1, 0.1, 0.025, 0.025])
        param_class = PowerLawFixedShear(self.kwargs_epl, 0.12)
        bounds = param_class.bounds(re_optimize=False)
        npt.assert_almost_equal(bounds[0], np.array(args) - shift)
        npt.assert_almost_equal(bounds[1], np.array(args) + shift)

        shift = np.array([0.005, 0.025, 0.025, 0.05, 0.05, 0.01, 0.01])
        param_class = PowerLawFixedShear(self.kwargs_epl, 0.12)
        bounds = param_class.bounds(re_optimize=True)
        npt.assert_almost_equal(bounds[0], np.array(args) - shift)
        npt.assert_almost_equal(bounds[1], np.array(args) + shift)

        scale = 0.7
        shift = scale * shift
        param_class = PowerLawFixedShear(self.kwargs_epl, 0.12)
        bounds = param_class.bounds(re_optimize=True, scale=scale)
        npt.assert_almost_equal(bounds[0], np.array(args) - shift)
        npt.assert_almost_equal(bounds[1], np.array(args) + shift)


if __name__ == "__main__":
    pytest.main()



"""Created on Dec 2, 2013.

@author: jakeret
"""
import py
import subprocess


def test_linkcheck(tmpdir):
    doctrees = tmpdir.join("doctrees")
    htmldir = tmpdir.join("html")
    subprocess.check_call(
        ["sphinx-build", "-blinkcheck", "-d", str(doctrees), ".", str(htmldir)]
    )


def test_build_docs(tmpdir):
    doctrees = tmpdir.join("doctrees")
    htmldir = tmpdir.join("html")
    subprocess.check_call(
        ["sphinx-build", "-bhtml", "-d", str(doctrees), ".", str(htmldir)]
    )



__author__ = "dgilman"

# this file contains a class to compute the Navaro-Frenk-White profile
import numpy as np
from scipy.special import hyp2f1
from scipy.special import beta
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["GNFW"]


class GNFW(LensProfileBase):
    """This class contains a double power law profile with flexible inner and outer
    logarithmic slopes g and n.

    .. math::
        \\rho(r) = \\frac{\\rho_0}{r^{\\gamma}} \\frac{Rs^{n}}{\\left(r^2 + Rs^2 \\right)^{(n - \\gamma)/2}}

    For g = 1.0 and n=3, it is approximately the same as an NFW profile
    The original reference is [1]_.

    .. [1] Munoz, Kochanek and Keeton, (2001), astro-ph/0103009, doi:10.1086/322314

    TODO: implement the gravitational potential for this profile
    """

    profile_name = "GNFW"
    param_names = [
        "Rs",
        "alpha_Rs",
        "center_x",
        "center_y",
        "gamma_inner",
        "gamma_outer",
    ]
    lower_limit_default = {
        "Rs": 0,
        "alpha_Rs": 0,
        "center_x": -100,
        "center_y": -100,
        "gamma_inner": 0.1,
        "gamma_outer": 1.0,
    }
    upper_limit_default = {
        "Rs": 100,
        "alpha_Rs": 10,
        "center_x": 100,
        "center_y": 100,
        "gamma_inner": 2.9,
        "gamma_outer": 10.0,
    }

    def derivatives(
        self, x, y, Rs, alpha_Rs, gamma_inner, gamma_outer, center_x=0, center_y=0
    ):
        """Returns df/dx and df/dy of the function which are the deflection angles.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param Rs: turn over point in the slope of the NFW profile in angular unit
        :param alpha_Rs: deflection (angular units) at projected Rs
        :param gamma_inner: logarithmic profile slope interior to Rs
        :param gamma_outer: logarithmic profile slope outside Rs
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: deflection angle in x, deflection angle in y
        """
        rho0_input = self.alpha2rho0(alpha_Rs, Rs, gamma_inner, gamma_outer)
        Rs = np.maximum(Rs, 0.00000001)
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        f_x, f_y = self.nfwAlpha(R, Rs, rho0_input, gamma_inner, gamma_outer, x_, y_)
        return f_x, f_y

    def hessian(
        self, x, y, Rs, alpha_Rs, gamma_inner, gamma_outer, center_x=0, center_y=0
    ):
        """

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param Rs: turn over point in the slope of the NFW profile in angular unit
        :param alpha_Rs: deflection (angular units) at projected Rs
        :param gamma_inner: logarithmic profile slope interior to Rs
        :param gamma_outer: logarithmic profile slope outside Rs
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx, d^f/dy^2
        """
        rho0_input = self.alpha2rho0(alpha_Rs, Rs, gamma_inner, gamma_outer)
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        R = np.maximum(R, 0.00000001)
        kappa = self.density_2d(R, 0, Rs, rho0_input, gamma_inner, gamma_outer)
        gamma1, gamma2 = self.nfwGamma(
            R, Rs, rho0_input, gamma_inner, gamma_outer, x_, y_
        )
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_xy, f_xy, f_yy

    @staticmethod
    def density(R, Rs, rho0, gamma_inner, gamma_outer):
        """Three dimensional NFW profile.

        :param R: radius of interest
        :type Rs: scale radius
        :param rho0: central density normalization
        :param gamma_inner: logarithmic profile slope interior to Rs
        :param gamma_outer: logarithmic profile slope outside Rs
        :return: rho(R) density
        """
        x = R / Rs
        outer_slope = (gamma_outer - gamma_inner) / 2
        return rho0 / (x**gamma_inner * (1 + x**2) ** outer_slope)

    def density_lens(self, r, Rs, alpha_Rs, gamma_inner, gamma_outer):
        """Computes the density at 3d radius r given lens model parameterization. The
        integral in the LOS projection of this quantity results in the convergence
        quantity.

        :param r: 3d radios
        :param Rs: scale radius
        :param alpha_Rs: deflection at Rs
        :param gamma_inner: logarithmic profile slope interior to Rs
        :param gamma_outer: logarithmic profile slope outside Rs
        :return: density rho(r)
        """
        rho0 = self.alpha2rho0(alpha_Rs, Rs, gamma_inner, gamma_outer)
        return self.density(r, Rs, rho0, gamma_inner, gamma_outer)

    def density_2d(
        self, x, y, Rs, rho0, gamma_inner, gamma_outer, center_x=0, center_y=0
    ):
        """Projected two dimensional profile.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param Rs: turn over point in the slope of the NFW profile in angular unit
        :param rho0: density normalization at Rs
        :param gamma_inner: logarithmic profile slope interior to Rs
        :param gamma_outer: logarithmic profile slope outside Rs
        :param center_x: profile center (same units as x)
        :param center_y: profile center (same units as x)
        :return: Epsilon(R) projected density at radius R
        """
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        x = R / Rs
        Fx = self._f(x, gamma_inner, gamma_outer)
        return 2 * rho0 * Rs * Fx

    @staticmethod
    def mass_3d(r, Rs, rho0, gamma_inner, gamma_outer):
        """Mass enclosed a 3d sphere or radius r.

        :param r: 3d radius
        :param Rs: scale radius
        :param rho0: density normalization
        :param gamma_inner: logarithmic profile slope interior to Rs
        :param gamma_outer: logarithmic profile slope outside Rs
        :return: M(<r)
        """
        Rs = float(Rs)
        const = 4 * np.pi * r**3 * rho0 * (Rs / r) ** gamma_inner
        m_3d = (
            const
            / (3 - gamma_inner)
            * hyp2f1(
                (3 - gamma_inner) / 2,
                (gamma_outer - gamma_inner) / 2,
                (5 - gamma_inner) / 2,
                -((r / Rs) ** 2),
            )
        )
        return m_3d

    def mass_3d_lens(self, r, Rs, alpha_Rs, gamma_inner, gamma_outer):
        """Mass enclosed a 3d sphere or radius r. This function takes as input the
        lensing parameterization.

        :param r: 3d radius
        :param Rs: scale radius
        :param alpha_Rs: deflection angle at Rs
        :param gamma_inner: logarithmic profile slope interior to Rs
        :param gamma_outer: logarithmic profile slope outside Rs
        :return: M(<r)
        """
        rho0 = self.alpha2rho0(alpha_Rs, Rs, gamma_inner, gamma_outer)
        m_3d = self.mass_3d(r, Rs, rho0, gamma_inner, gamma_outer)
        return m_3d

    def mass_2d(self, R, Rs, rho0, gamma_inner, gamma_outer):
        """Mass enclosed a 2d cylinder or projected radius R.

        :param R: 3d radius
        :param Rs: scale radius
        :param rho0: central density normalization
        :param gamma_inner: logarithmic profile slope interior to Rs
        :param gamma_outer: logarithmic profile slope outside Rs
        :return: mass in cylinder
        """
        R = np.maximum(R, 0.00000001)
        x = R / Rs
        gx = self._g(x, gamma_inner, gamma_outer)
        m_2d = 4 * rho0 * Rs * R**2 * gx / x**2 * np.pi
        return m_2d

    def nfwAlpha(self, R, Rs, rho0, gamma_inner, gamma_outer, ax_x, ax_y):
        """Deflection angel of NFW profile (times Sigma_crit D_OL) along the projection
        to coordinate 'axis'.

        :param R: 3d radius
        :param Rs: scale radius
        :param rho0: central density normalization
        :param gamma_inner: logarithmic profile slope interior to Rs
        :param gamma_outer: logarithmic profile slope outside Rs
        :param ax_x: x coordinate relative to center
        :param ax_y: y coordinate relative to center
        :return: Epsilon(R) projected density at radius R
        """
        R = np.maximum(R, 0.00000001)
        x = R / Rs
        gx = self._g(x, gamma_inner, gamma_outer)
        a = 4 * rho0 * Rs * R * gx / x**2 / R
        return a * ax_x, a * ax_y

    def nfwGamma(self, R, Rs, rho0, gamma_inner, gamma_outer, ax_x, ax_y):
        """Shear gamma of NFW profile (times Sigma_crit) along the projection to
        coordinate 'axis'.

        :param R: 3d radius
        :param Rs: scale radius
        :param rho0: central density normalization
        :param gamma_inner: logarithmic profile slope interior to Rs
        :param gamma_outer: logarithmic profile slope outside Rs
        :param ax_x: x coordinate relative to center
        :param ax_y: y coordinate relative to center
        :return: Epsilon(R) projected density at radius R
        """
        R = np.maximum(R, 0.00000001)
        x = R / Rs
        gx = self._g(x, gamma_inner, gamma_outer)
        Fx = self._f(x, gamma_inner, gamma_outer)
        a = (
            2 * rho0 * Rs * (2 * gx / x**2 - Fx)
        )  # /x #2*rho0*Rs*(2*gx/x**2 - Fx)*axis/x
        return a * (ax_y**2 - ax_x**2) / R**2, -a * 2 * (ax_x * ax_y) / R**2

    @staticmethod
    def _f(X, g, n):
        """Analytic solution of the projection integral.

        :param X: R/Rs
        :type X: float >0
        :param g: logarithmic profile slope interior to Rs
        :param n: logarithmic profile slope exterior to Rs
        :return: solution to the projection integral
        """
        if n == 3:
            n = 3.001  # for numerical stability
        hyp2f1_term = hyp2f1((n - 1) / 2, g / 2, n / 2, 1 / (1 + X**2))
        beta_term = beta((n - 1) / 2, 0.5)
        return 0.5 * beta_term * hyp2f1_term * (1 + X**2) ** ((1 - n) / 2)

    @staticmethod
    def _g(X, g, n):
        """Analytic solution of integral for NFW profile to compute deflection angel and
        gamma.

        :param X: R/Rs
        :type X: float >0
        :param g: logarithmic profile slope interior to Rs
        :param n: logarithmic profile slope exterior to Rs
        :return: solution of the integral over projected mass
        """
        if n == 3:
            n = 3.001  # for numerical stability
        xi = 1 + X**2
        hyp2f1_term = hyp2f1((n - 3) / 2, g / 2, n / 2, 1 / xi)
        beta_term_1 = beta((n - 3) / 2, (3 - g) / 2)
        beta_term_2 = beta((n - 3) / 2, 1.5)
        return 0.5 * (beta_term_1 - beta_term_2 * hyp2f1_term * xi ** ((3 - n) / 2))

    def alpha2rho0(self, alpha_Rs, Rs, gamma_inner, gamma_outer):
        """Convert angle at Rs into rho0.

        :param alpha_Rs: deflection angle at RS
        :param Rs: scale radius
        :param gamma_inner: logarithmic profile slope interior to Rs
        :param gamma_outer: logarithmic profile slope outside Rs
        :return: density normalization (characteristic density)
        """

        gx = self._g(1.0, gamma_inner, gamma_outer)
        rho0 = alpha_Rs / (4.0 * Rs**2 * gx / 1.0**2)
        return rho0

    def rho02alpha(self, rho0, Rs, gamma_inner, gamma_outer):
        """Convert rho0 to angle at Rs.

        :param rho0: density normalization (characteristic density)
        :param Rs: scale radius
        :param gamma_inner: logarithmic profile slope interior to Rs
        :param gamma_outer: logarithmic profile slope outside Rs
        :return: deflection angle at RS
        """
        gx = self._g(1.0, gamma_inner, gamma_outer)
        alpha_Rs = rho0 * (4.0 * Rs**2 * gx / 1.0**2)
        return alpha_Rs



__author__ = "aymgal"

import pytest
import numpy as np
import numpy.testing as npt

from lenstronomy.Sampling.Samplers.dynesty_sampler import DynestySampler


@pytest.fixture
def import_fixture(simple_einstein_ring_likelihood):
    """

    :param simple_einstein_ring_likelihood: fixture
    :return:
    """
    likelihood, kwargs_truths = simple_einstein_ring_likelihood
    prior_means = likelihood.param.kwargs2args(**kwargs_truths)
    prior_means *= 1.01
    prior_sigmas = np.ones_like(prior_means)
    print(prior_sigmas, prior_means, "test prior sigmas")
    sampler = DynestySampler(
        likelihood,
        prior_type="uniform",
        prior_means=prior_means,
        prior_sigmas=prior_sigmas,
        sigma_scale=0.5,
    )
    return sampler, likelihood


class TestDynestySampler(object):
    """Test the fitting sequences."""

    def setup_method(self):
        pass

    def test_sampler(self, import_fixture):
        sampler, likelihood = import_fixture
        kwargs_run = {
            "dlogz_init": 0.01,
            "nlive_init": 20,
            "nlive_batch": 20,
            "maxbatch": 1,
            "wt_kwargs": {"pfrac": 0.8},
        }
        samples, means, logZ, logZ_err, logL, results = sampler.run(kwargs_run)
        assert len(means) == 1

    def test_sampler_init(self, import_fixture):
        sampler, likelihood = import_fixture
        try:
            sampler = DynestySampler(
                likelihood,
                prior_type="gaussian",
                prior_means=None,  # will raise an Error
                prior_sigmas=None,
            )  # will raise an Error
        except Exception as e:
            assert isinstance(e, ValueError)
        try:
            sampler = DynestySampler(likelihood, prior_type="some_type")
        except Exception as e:
            assert isinstance(e, ValueError)

    def test_prior(self, import_fixture):
        sampler, likelihood = import_fixture
        n_dims = sampler.n_dims
        cube_low = np.zeros(n_dims)
        cube_upp = np.ones(n_dims)

        self.prior_type = "uniform"
        cube_low = sampler.prior(cube_low)
        npt.assert_equal(cube_low, sampler.lowers)
        cube_upp = sampler.prior(cube_upp)
        npt.assert_equal(cube_upp, sampler.uppers)

    def test_log_likelihood(self, import_fixture):
        sampler, likelihood = import_fixture
        n_dims = sampler.n_dims
        args = np.nan * np.ones(n_dims)
        logL = sampler.log_likelihood(args)
        assert logL < 0
        # npt.assert_almost_equal(logL, -47.167446538898204)
        # assert logL == -1e15


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
import lenstronomy.Util.util as util
from lenstronomy.ImSim.image2source_mapping import Image2SourceMapping
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
import pytest
import unittest


class TestMultiSourcePlane(object):
    def setup_method(self):
        lens_model_list = ["SIS", "SIS"]
        self.kwargs_lens = [
            {"theta_E": 1, "center_x": 0, "center_y": 0},
            {"theta_E": 0.5, "center_x": 1, "center_y": 1},
        ]
        singlePlane = LensModel(lens_model_list=lens_model_list)
        multiPlane = LensModel(
            lens_model_list=lens_model_list,
            multi_plane=True,
            z_source=3,
            lens_redshift_list=[0.2, 0.5],
            cosmo=None,
        )
        pseudoMultiPlane = LensModel(
            lens_model_list=lens_model_list,
            multi_plane=True,
            z_source=3,
            lens_redshift_list=[0.5, 0.5],
            cosmo=None,
        )
        # test single plane single source

        # test single plane multi source

        # test pseudo multi plane single source

        light_model_list = ["SERSIC", "SERSIC"]
        self.kwargs_light = [
            {"amp": 1, "R_sersic": 1, "n_sersic": 2, "center_x": 0, "center_y": 0},
            {"amp": 2, "R_sersic": 0.5, "n_sersic": 1, "center_x": 1, "center_y": 1},
        ]
        self.singlePlane_singlePlane = Image2SourceMapping(
            singlePlane,
            LightModel(
                light_model_list,
                deflection_scaling_list=None,
                source_redshift_list=None,
            ),
        )
        self.singlePlane_pseudoMulti = Image2SourceMapping(
            singlePlane,
            LightModel(
                light_model_list,
                deflection_scaling_list=[1, 1],
                source_redshift_list=None,
            ),
        )
        self.pseudoMulti_pseudoMulti = Image2SourceMapping(
            pseudoMultiPlane,
            LightModel(
                light_model_list,
                deflection_scaling_list=None,
                source_redshift_list=[3, 3],
            ),
        )
        self.pseudoMulti_single = Image2SourceMapping(
            pseudoMultiPlane,
            LightModel(
                light_model_list,
                deflection_scaling_list=None,
                source_redshift_list=None,
            ),
        )
        self.multi_single = Image2SourceMapping(
            multiPlane,
            LightModel(
                light_model_list,
                deflection_scaling_list=None,
                source_redshift_list=None,
            ),
        )
        self.multi_pseudoMulti = Image2SourceMapping(
            multiPlane,
            LightModel(
                light_model_list,
                deflection_scaling_list=None,
                source_redshift_list=[3, 3],
            ),
        )
        self.multi_multi = Image2SourceMapping(
            multiPlane,
            LightModel(
                light_model_list,
                deflection_scaling_list=None,
                source_redshift_list=[0.3, 2],
            ),
        )

    def test_pseudo_multi_ray_tracing(self):
        x, y = util.make_grid(numPix=10, deltapix=0.5)
        kwargs_lens = self.kwargs_lens
        kwargs_light = self.kwargs_light
        flux_single_single = self.singlePlane_singlePlane.image_flux_joint(
            x, y, kwargs_lens=kwargs_lens, kwargs_source=kwargs_light
        )
        flux_single_pseudo = self.singlePlane_pseudoMulti.image_flux_joint(
            x, y, kwargs_lens=kwargs_lens, kwargs_source=kwargs_light
        )
        flux_pseudo_pseudo = self.pseudoMulti_pseudoMulti.image_flux_joint(
            x, y, kwargs_lens=kwargs_lens, kwargs_source=kwargs_light
        )
        flux_pseudo_single = self.pseudoMulti_single.image_flux_joint(
            x, y, kwargs_lens=kwargs_lens, kwargs_source=kwargs_light
        )
        npt.assert_almost_equal(flux_single_single, flux_single_pseudo, decimal=10)
        npt.assert_almost_equal(flux_single_single, flux_pseudo_pseudo, decimal=10)
        npt.assert_almost_equal(flux_single_single, flux_pseudo_single, decimal=10)

    def test_multi_ray_tracing(self):
        x, y = util.make_grid(numPix=10, deltapix=0.1)
        kwargs_lens = self.kwargs_lens
        kwargs_light = self.kwargs_light
        flux_multi_single = self.multi_single.image_flux_joint(
            x, y, kwargs_lens=kwargs_lens, kwargs_source=kwargs_light
        )
        flux_multi_pseudo = self.multi_pseudoMulti.image_flux_joint(
            x, y, kwargs_lens=kwargs_lens, kwargs_source=kwargs_light
        )
        npt.assert_almost_equal(flux_multi_pseudo, flux_multi_single, decimal=10)

        flux_multi_multi = self.multi_multi.image_flux_joint(
            x, y, kwargs_lens=kwargs_lens, kwargs_source=kwargs_light
        )

        # import matplotlib.pyplot as plt
        # plt.matshow(util.array2image(flux_multi_multi))
        # plt.show()
        npt.assert_almost_equal(np.sum(flux_multi_multi), 1454.689246553742, decimal=-1)

    def test_pseudo_ray_trace_functions_split(self):
        x, y = util.make_grid(numPix=10, deltapix=0.5)
        kwargs_lens = self.kwargs_lens
        kwargs_light = self.kwargs_light
        response_single_single, n1 = self.singlePlane_singlePlane.image_flux_split(
            x, y, kwargs_lens=kwargs_lens, kwargs_source=kwargs_light
        )
        response_single_pseudo, n2 = self.singlePlane_pseudoMulti.image_flux_split(
            x, y, kwargs_lens=kwargs_lens, kwargs_source=kwargs_light
        )
        response_pseudo_pseudo, n3 = self.pseudoMulti_pseudoMulti.image_flux_split(
            x, y, kwargs_lens=kwargs_lens, kwargs_source=kwargs_light
        )
        response_pseudo_single, n4 = self.pseudoMulti_single.image_flux_split(
            x, y, kwargs_lens=kwargs_lens, kwargs_source=kwargs_light
        )
        npt.assert_almost_equal(n1, n2, decimal=10)
        npt.assert_almost_equal(n1, n3, decimal=10)
        npt.assert_almost_equal(n1, n4, decimal=10)
        assert n1 == 2
        npt.assert_almost_equal(
            response_single_single[0], response_single_pseudo[0], decimal=10
        )
        npt.assert_almost_equal(
            response_single_single[0], response_pseudo_pseudo[0], decimal=10
        )
        npt.assert_almost_equal(
            response_single_single[0], response_pseudo_single[0], decimal=10
        )

        npt.assert_almost_equal(
            response_single_single[1], response_single_pseudo[1], decimal=10
        )
        npt.assert_almost_equal(
            response_single_single[1], response_pseudo_pseudo[1], decimal=10
        )
        npt.assert_almost_equal(
            response_single_single[1], response_pseudo_single[1], decimal=10
        )

    def test_multi_ray_trace_functions_split(self):
        x, y = util.make_grid(numPix=10, deltapix=0.1)
        kwargs_lens = self.kwargs_lens
        kwargs_light = self.kwargs_light
        response_multi_single, n1 = self.multi_single.image_flux_split(
            x, y, kwargs_lens=kwargs_lens, kwargs_source=kwargs_light
        )
        response_multi_pseudo, n2 = self.multi_pseudoMulti.image_flux_split(
            x, y, kwargs_lens=kwargs_lens, kwargs_source=kwargs_light
        )
        npt.assert_almost_equal(
            response_multi_pseudo[0], response_multi_single[0], decimal=10
        )
        npt.assert_almost_equal(
            response_multi_pseudo[1], response_multi_single[1], decimal=10
        )
        npt.assert_almost_equal(n1, n2, decimal=10)
        assert n1 == 2

        response_multi_multi, n = self.multi_multi.image_flux_split(
            x, y, kwargs_lens=kwargs_lens, kwargs_source=kwargs_light
        )
        npt.assert_almost_equal(np.sum(response_multi_multi), 1413, decimal=-1)

    def test_image2source(self):
        x, y = 1, 1
        beta_x, beta_y = self.multi_multi.image2source(
            x, y, kwargs_lens=self.kwargs_lens, index_source=0
        )
        npt.assert_almost_equal(beta_x, 0.7433428403740511, decimal=2)

        beta_x0, beta_y0 = self.singlePlane_singlePlane.image2source(
            x, y, kwargs_lens=self.kwargs_lens, index_source=0
        )
        beta_x, beta_y = self.singlePlane_pseudoMulti.image2source(
            x, y, kwargs_lens=self.kwargs_lens, index_source=0
        )
        npt.assert_almost_equal(beta_x0, beta_x, decimal=10)
        beta_x, beta_y = self.pseudoMulti_pseudoMulti.image2source(
            x, y, kwargs_lens=self.kwargs_lens, index_source=0
        )
        npt.assert_almost_equal(beta_x0, beta_x, decimal=10)

    def test__re_order_split(self):
        lensModel = LensModel(
            lens_model_list=["SIS", "SIS"],
            multi_plane=True,
            lens_redshift_list=[0.5, 0.4],
            z_source=3,
        )
        mapping = Image2SourceMapping(
            lensModel,
            LightModel(
                light_model_list=["SERSIC", "SHAPELETS"],
                deflection_scaling_list=None,
                source_redshift_list=[2, 0.3],
            ),
        )
        n_list = [1, 2]
        response = np.zeros((3, 3))
        response[1:] = 1
        response_reshuffled = mapping._re_order_split(response, n_list)
        assert response_reshuffled[0, 0] == 1
        assert response_reshuffled[1, 0] == 0


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            lensModel = LensModel(
                lens_model_list=["SIS"],
                multi_plane=True,
                z_source=3,
                lens_redshift_list=[0.2],
                cosmo=None,
            )
            lightModel = LightModel(
                light_model_list=["UNIFORM"],
                deflection_scaling_list=[1.0],
                source_redshift_list=None,
            )
            class_instance = Image2SourceMapping(lensModel, lightModel)

        with self.assertRaises(ValueError):
            lensModel = LensModel(
                lens_model_list=["SIS"],
                multi_plane=True,
                z_source=3,
                lens_redshift_list=[0.2],
                cosmo=None,
            )
            lightModel = LightModel(
                light_model_list=["UNIFORM"],
                deflection_scaling_list=None,
                source_redshift_list=[0, 1, 2],
            )
            class_instance = Image2SourceMapping(lensModel, lightModel)

        with self.assertRaises(ValueError):
            lensModel = LensModel(
                lens_model_list=["SIS"],
                multi_plane=True,
                z_source=0.5,
                lens_redshift_list=[0.2],
                cosmo=None,
            )
            lightModel = LightModel(
                light_model_list=["UNIFORM"],
                deflection_scaling_list=None,
                source_redshift_list=[1],
            )
            class_instance = Image2SourceMapping(lensModel, lightModel)

        with self.assertRaises(ValueError):
            lensModel = LensModel(
                lens_model_list=["SIS"], multi_plane=False, z_source=0.5, cosmo=None
            )
            lightModel = LightModel(
                light_model_list=["UNIFORM"], deflection_scaling_list=[1, 1]
            )
            class_instance = Image2SourceMapping(lensModel, lightModel)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"


from lenstronomy.LensModel.Profiles.point_mass import PointMass

import numpy as np
import numpy.testing as npt
import pytest


class TestSIS(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.pointmass = PointMass()

    def test_function(self):
        x = np.array([0])
        y = np.array([1])
        theta_E = 1.0
        values = self.pointmass.function(x, y, theta_E)
        assert values[0] == 0
        x = np.array([0])
        y = np.array([0])
        values = self.pointmass.function(x, y, theta_E)
        assert values[0] < 0

        x = np.array([1, 3, 4])
        y = np.array([0, 1, 1])
        values = self.pointmass.function(x, y, theta_E)
        assert values[0] == 0
        assert values[1] == 1.151292546497023
        assert values[2] == 1.4166066720281081

    def test_derivatives(self):
        x = np.array([1])
        y = np.array([0])
        theta_E = 1.0
        f_x, f_y = self.pointmass.derivatives(x, y, theta_E)
        assert f_x[0] == 1
        assert f_y[0] == 0
        x = np.array([0])
        y = np.array([0])
        f_x, f_y = self.pointmass.derivatives(x, y, theta_E)
        assert f_x[0] == 0
        assert f_y[0] == 0

        x = np.array([1, 3, 4])
        y = np.array([0, 1, 1])
        values = self.pointmass.derivatives(x, y, theta_E)
        assert values[0][0] == 1
        assert values[1][0] == 0
        assert values[0][1] == 0.29999999999999999
        assert values[1][1] == 0.099999999999999992

    def test_hessian(self):
        x = np.array([1])
        y = np.array([0])
        theta_E = 1.0
        f_xx, f_xy, f_yx, f_yy = self.pointmass.hessian(x, y, theta_E)
        npt.assert_almost_equal(f_xy, f_yx, decimal=8)
        assert f_xx[0] == -1
        assert f_yy[0] == 1
        assert f_xy[0] == -0
        x = np.array([1, 3, 4])
        y = np.array([0, 1, 1])
        values = self.pointmass.hessian(x, y, theta_E)
        assert values[0][0] == -1
        assert values[3][0] == 1
        assert values[1][0] == -0
        assert values[0][1] == -0.080000000000000002
        assert values[3][1] == 0.080000000000000002
        assert values[1][1] == -0.059999999999999998


if __name__ == "__main__":
    pytest.main()



import numpy as np
from scipy import stats

__all__ = ["KDELikelihood"]


class KDELikelihood(object):
    """Class that samples the cosmographic likelihood given a distribution of points in
    the 2-dimensional distribution of D_d and D_delta_t."""

    def __init__(
        self, D_d_sample, D_delta_t_sample, kde_type="scipy_gaussian", bandwidth=1
    ):
        """

        :param D_d_sample: 1-d numpy array of angular diameter distances to the lens plane
        :param D_delta_t_sample: 1-d numpy array of time-delay distances
        :param kde_type: The kernel to use.  Valid kernels are
         'scipy_gaussian' or
         ['gaussian'|'tophat'|'epanechnikov'|'exponential'|'linear'|'cosine']
         Default is 'gaussian'.
        :type kde_type: string
        :param bandwidth: width of kernel (in same units as the angular diameter quantities)
        """
        values = np.vstack([D_d_sample, D_delta_t_sample])
        if kde_type == "scipy_gaussian":
            self._PDF_kernel = stats.gaussian_kde(values)
        else:
            from sklearn.neighbors import KernelDensity

            self._kde = KernelDensity(bandwidth=bandwidth, kernel=kde_type)
            values = np.vstack([D_d_sample, D_delta_t_sample])
            self._kde.fit(values.T)
        self._kde_type = kde_type

    def logLikelihood(self, D_d, D_delta_t):
        """Likelihood of the data (represented in the distribution of this class) given
        a model with predicted angular diameter distances.

        :param D_d: model predicted angular diameter distance
        :param D_delta_t: model predicted time-delay distance
        :return: loglikelihood (log of KDE value)
        """
        if self._kde_type == "scipy_gaussian":
            density = self._PDF_kernel([D_d, D_delta_t])
            logL = np.log(density)
        else:
            x = np.array([[D_d], [D_delta_t]])
            logL = self._kde.score_samples(x.T)
        return logL
